from .config import xp as np
from .frequency_severity import FreqSevSims
from dataclasses import dataclass
from .variables import StochasticScalar


@dataclass
class ContractResults:
    """A class to hold the recoveries and reinstatement premiums resulting from applying a reinsurance contract to a
    set of claims."""

    def __init__(
        self,
        recoveries: FreqSevSims,
        reinstatement_premium: StochasticScalar | None = None,
    ):
        """
        Create a new contract results object.

        Args:
            recoveries (FreqSevSims): Object representing the recoveries.
            reinstatement_premium (np.ndarray | None): Optional Array representing the reinstatement premium.
        """
        self.recoveries = recoveries
        self.reinstatement_premium = reinstatement_premium


class XoL:
    """Represents an excess of loss reinsurance contract"""

    def __init__(
        self,
        name: str,
        limit: float,
        excess: float,
        premium: float,
        reinstatement_cost: list[float] | None = None,
        aggregate_limit: float | None = None,
        aggregate_deductible: float | None = None,
        franchise: float | None = None,
        reverse_franchise: float | None = None,
    ):
        """
        Initialize a new XoL layer.

        Args:
            name (str): The name of the XoL layer.
            limit (float): The limit of coverage.
            excess (float): The excess amount.
            premium (float): The premium amount.
            reinstatement_cost (list[float] | None, optional): The reinstatement cost as a fraction of the base
                premium, by reinstatement. Defaults to None.
            aggregate_limit (float, optional): The aggregate limit. Defaults to None.
            aggregate_deductible (float, optional): The aggregate deductible. Defaults to None.
            franchise (float, optional): The franchise amount. Defaults to None.
            reverse_franchise (float, optional): The reverse franchise amount. Defaults to None.
        """
        self.name = name
        self.limit = limit
        self.excess = excess
        self.aggregate_limit: float = (
            aggregate_limit if aggregate_limit is not None else np.inf
        )
        self.premium = premium
        self.aggregate_deductible: float = (
            aggregate_deductible if aggregate_deductible is not None else 0
        )
        self.franchise: float = franchise if franchise is not None else 0
        self.reverse_franchise: float = (
            reverse_franchise if reverse_franchise is not None else np.inf
        )
        self.summary: dict[str, float] = {}
        self.num_reinstatements = (
            aggregate_limit / limit - 1 if aggregate_limit is not None else None
        )
        self.reinstatement_premium_cost = (
            np.array(reinstatement_cost) if reinstatement_cost is not None else None
        )

    def apply(self, claims: FreqSevSims) -> ContractResults:
        """
        Apply the XoL contract to a set of claims.

        Args:
            claims (FreqSevSims): The simulated claims to apply the contract to.

        Returns:
            ContractResults: The results of applying the contract.

        Calculation of the recoveries from the excess of loss contract:

        Firstly, the effect of any franchise or reverse franchise is calculated on the individual losses.

        losses post franchise = loss if loss >= franchise and loss<= reverse franchise

        Next the individual losses to the layer are calculated:

        layer_loss = min(max(losses post franchise - excess, 0), limit)

        Then the aggregate layer losses before aggregate limit and deductible are calculated.
        The aggregate limit and deductible are then applied to get the aggregate recoveries for the layer:

        aggregate_recoveries = min(max(aggregate_layer_losses - aggregate_deductible, 0), aggregate_limit)

        The aggregate recoveries are then allocated back to the individual losses in proportion to the individual recoveries before aggregate limit and deductible.

        The reinstatement premium is calculated as the sum of the reinstatement premium cost multiplied by the number of reinstatements used.
        The number of reinstatements used is calculated as the minimum of the aggregate recoveries divided by the limit and the number of reinstatements available (which is the aggregate limit divided by the occurrence limit, less one).
        """
        # apply franchise
        if self.franchise != 0 or self.reverse_franchise != np.inf:
            is_in_window = (claims >= self.franchise) & (
                claims < self.reverse_franchise
            )
            claims = np.where(
                is_in_window,
                claims,
                0,
            )

        individual_recoveries_pre_aggregate: FreqSevSims = np.minimum(
            np.maximum(claims - self.excess, 0), self.limit
        )
        if self.aggregate_limit == np.inf and self.aggregate_deductible == 0:
            self.calc_summary(claims, individual_recoveries_pre_aggregate.aggregate())
            return ContractResults(individual_recoveries_pre_aggregate)
        aggregate_limit = (
            self.aggregate_limit if self.aggregate_limit is not None else np.inf
        )
        aggregate_deductible = (
            self.aggregate_deductible if self.aggregate_deductible is not None else 0
        )
        aggregate_recoveries_pre_agg = individual_recoveries_pre_aggregate.aggregate()

        aggregate_recoveries = np.minimum(
            np.maximum(aggregate_recoveries_pre_agg - aggregate_deductible, 0),
            aggregate_limit,
        )
        non_zero_recoveries = aggregate_recoveries != 0
        ratio = np.ones(aggregate_recoveries_pre_agg.values.shape)
        np.putmask(
            ratio,
            non_zero_recoveries.values,
            np.divide(
                aggregate_recoveries.values[non_zero_recoveries.values],
                aggregate_recoveries_pre_agg.values[non_zero_recoveries.values],
            ),
        )

        recoveries = individual_recoveries_pre_aggregate * ratio
        results = ContractResults(recoveries)
        if (
            self.reinstatement_premium_cost is not None
            and self.num_reinstatements is not None
        ):
            cumulative_reinstatement_cost = np.cumsum(self.reinstatement_premium_cost)
            limits_used = aggregate_recoveries / self.limit
            reinstatements_used = np.minimum(limits_used, self.num_reinstatements)
            reinstatements_used_full = StochasticScalar(
                np.floor(reinstatements_used).values.astype(int)
            )
            reinstatements_used_fraction = (
                reinstatements_used - reinstatements_used_full
            )
            reinstatement_number = np.maximum(reinstatements_used_full - 1, 0)
            reinstatement_premium_proportion = self.reinstatement_premium_cost[
                reinstatement_number.values
            ] * reinstatements_used_fraction + np.where(
                reinstatements_used_full.values > 0,
                cumulative_reinstatement_cost[reinstatement_number.values],
                0,
            )
            reinstatement_premium = reinstatement_premium_proportion * self.premium
            results.reinstatement_premium = reinstatement_premium
        self.calc_summary(claims, aggregate_recoveries)
        return results

    def calc_summary(
        self, gross_losses: FreqSevSims, aggregate_recoveries: StochasticScalar
    ):
        """
        Calculate a summary of the losses to the layer. The results are stored in the summary attribute of the layer.

        The summary includes the mean and standard deviation of the recoveries, the probability of attachment, the probability of vertical exhaustion and the probability of horizontal exhaustion.

        Args:
            gross_losses (FreqSevSims): Object representing the gross losses.
            aggregate_recoveries (StochasticScalar): Array of aggregate recoveries.

        Returns:
            None

        """
        mean = aggregate_recoveries.mean()
        sd = aggregate_recoveries.std()
        count = (aggregate_recoveries > 0).ssum()
        vertical_exhaust: FreqSevSims = np.maximum(
            gross_losses - self.limit + self.excess, 0
        )
        aggregate_vertical_exhaust = vertical_exhaust.aggregate()

        v_count = (aggregate_vertical_exhaust > 0).ssum()
        h_count = (aggregate_recoveries >= self.aggregate_limit).ssum()

        self.summary = {
            "mean": mean,
            "std": sd,
            "prob_attach": count / aggregate_recoveries.n_sims,
            "prob_vert_exhaust": v_count / len(gross_losses.values),
            "prob_horizonal_exhaust": (
                h_count / aggregate_recoveries.n_sims
                if self.aggregate_limit is not None
                else 0
            ),
        }

    def print_summary(self):
        """Print a summary of the losses to the layer

        >>> layer.print_summary()
        Layer Name : Layer 1
        Mean Recoveries:  100000.0
        SD Recoveries:  0.0
        Probability of Attachment:  0.0
        Probability of Vertical Exhaustion:  0.0
        Probability of Horizontal Exhaustion:  0.0

        """

        print("Layer Name : {}".format(self.name))
        print("Mean Recoveries: ", self.summary["mean"])
        print("SD Recoveries: ", self.summary["std"])
        print("Probability of Attachment: ", self.summary["prob_attach"]),

        print(
            "Probability of Vertical Exhaustion: ", self.summary["prob_vert_exhaust"]
        ),
        print(
            "Probability of Horizontal Exhaustion: ",
            self.summary["prob_horizonal_exhaust"],
        ),
        print("")


class XoLTower:
    """Represents a tower of excess of loss reinsurance contracts."""

    def __init__(
        self,
        limit: list[float],
        excess: list[float],
        premium: list[float],
        name: list[str] | None = None,
        reinstatement_cost: list[list[float] | None] | None = None,
        aggregate_deductible: list[float | None] | None = None,
        aggregate_limit: list[float | None] | None = None,
        franchise: list[float | None] | None = None,
        reverse_franchise: list[float | None] | None = None,
    ):
        """
        Create an XoL Tower.

        Args:
            limit (list[float]): A list of limits for each layer.
            excess (list[float]): A list of excesses for each layer.
            premium (list[float]): The premium for each layer.
            name (list[str], optional): A list of names for each layer. Defaults to None.
            reinstatement_cost (list[list[float]] | None, optional): A list of reinstatement costs for each reinstatement for each layer. Defaults to None.
            aggregate_deductible (list[float], optional): The aggregate deductible. Defaults to None.
            aggregate_limit (list[float], optional): The aggregate limit. Defaults to None.
            franchise (list[float], optional): The franchise amount. Defaults to None.
            reverse_franchise (list[float], optional): The reverse franchise amount. Defaults to None.
        """

        self.limit = limit
        self.excess = excess
        self.aggregate_limit = aggregate_limit
        self.aggregate_deductible = aggregate_deductible
        self.franchise = franchise
        self.reverse_franchise = reverse_franchise
        self.n_layers = len(limit)
        self.layers = [
            XoL(
                "Layer {}".format(i + 1) if name is None else name[i],
                limit[i],
                excess[i],
                premium[i],
                reinstatement_cost[i] if reinstatement_cost is not None else None,
                aggregate_limit[i] if aggregate_limit is not None else None,
                aggregate_deductible[i] if aggregate_deductible is not None else None,
                franchise[i] if franchise is not None else None,
                reverse_franchise[i] if reverse_franchise is not None else None,
            )
            for i in range(self.n_layers)
        ]

    def apply(self, claims: FreqSevSims) -> ContractResults:
        """
        Applies the XoL Tower to a set of claims.

        Parameters:
            claims (FreqSevSims): The set of claims to apply the XoL Tower to.

        Returns:
            ContractResults: The results of applying the XoL Tower to the claims.
        """
        recoveries = claims.copy() * 0
        reinstatement_premium = StochasticScalar(np.zeros(claims.n_sims))
        for layer in self.layers:
            layer_results = layer.apply(claims)
            recoveries += layer_results.recoveries
            reinstatement_premium += (
                layer_results.reinstatement_premium
                if layer_results.reinstatement_premium is not None
                else 0
            )

        return ContractResults(recoveries, reinstatement_premium)

    def print_summary(self):
        """Print a summary of the program losses"""
        for layer in self.layers:
            layer.print_summary()



from pal.variables import ProteusVariable as pv, StochasticScalar


def test_empty():
    x = pv(dim_name="dim1", values=[])
    assert x.values == []


def test_variable():
    x = pv(dim_name="dim1", values=[1, 2, 3])
    y = x + 1
    assert y.values == [2, 3, 4]


def test_variable2():
    """Test that a variable can be created with a list of StochasticScalars."""
    x = pv(
        dim_name="dim1",
        values=[StochasticScalar([1, 2, 3]), StochasticScalar([2, 3, 4])],
    )
    y = x + 2.2
    assert (
        y
        == pv(
            dim_name="dim1",
            values=[
                StochasticScalar([3.2, 4.2, 5.2]),
                StochasticScalar([4.2, 5.2, 6.2]),
            ],
        )
    ).all()


def test_variable3():
    """Test that a variable can be created with a dictionary, that label matching works, and that the variable can be summed."""
    x = pv(
        dim_name="dim1",
        values={"a": 1, "b": 2},
    )
    y = pv(
        dim_name="dim1",
        values={"b": 5, "a": 8},
    )
    z = x + y
    assert z.values == {"a": 9, "b": 7}


def test_dict_variable_dereferencing():
    x = pv(
        dim_name="dim1",
        values={"a": 1, "b": 2},
    )
    assert x["a"] == 1
    assert x["b"] == 2
    assert x[0] == 1
    assert x[1] == 2


def test_array_variable_dereferencing():
    x = pv(
        dim_name="dim1",
        values=[1, 2],
    )
    assert x[0] == 1
    assert x[1] == 2


def test_sum():
    x = pv(dim_name="dim1", values=[1, 2])
    y = sum(x)
    assert y == 3


def test_sum_stochastic():
    x = pv(
        dim_name="dim1",
        values=[StochasticScalar([1, 2, 3]), StochasticScalar([2, 3, 4])],
    )
    y = sum(x)
    assert (y == StochasticScalar([3, 5, 7])).all()


def test_sum_dict_stochastic():
    x = pv(
        dim_name="dim1",
        values={"a": StochasticScalar([1, 2, 3]), "b": StochasticScalar([2, 3, 4])},
    )
    y = sum(x)
    assert (y == StochasticScalar([3, 5, 7])).all()
    assert (
        y.coupled_variable_group
        == x[0].coupled_variable_group
        == x[1].coupled_variable_group
    )


import numpy as np


def test_corr():
    x = pv(
        dim_name="dim1",
        values={"a": StochasticScalar([1, 10, 2]), "b": StochasticScalar([2, 3, 4])},
    )
    matrix = x.correlation_matrix()
    assert (np.array(matrix) == np.array([[1, 0.5], [0.5, 1]])).all()


def test_get_value_at_sim():
    x = pv(
        dim_name="dim1",
        values={"a": StochasticScalar([1, 2, 3]), "b": StochasticScalar([2, 3, 4])},
    )
    assert x.get_value_at_sim(0).values == {"a": 1, "b": 2}
    assert x.get_value_at_sim(1).values == {"a": 2, "b": 3}


def test_get_value_at_sim_stochastic():
    x = pv(
        dim_name="dim1",
        values={"a": StochasticScalar([1, 2, 3]), "b": StochasticScalar([2, 3, 4])},
    )
    assert (
        x.get_value_at_sim(StochasticScalar([0, 2]))
        == pv("dim1", {"a": StochasticScalar([1, 3]), "b": StochasticScalar([2, 4])})
    ).all()


def test_array_ufunc():
    x = pv(
        dim_name="dim1",
        values=[StochasticScalar([1, 2, 3])],
    )
    y = np.exp(x)
    assert (y.values[0] == StochasticScalar([np.exp(1), np.exp(2), np.exp(3)])).all()


def test_array_func2():
    x = pv(
        dim_name="dim1",
        values=[StochasticScalar([1, 2, 3]), StochasticScalar([1, 2, 3])],
    )
    y = np.cumsum(x)
    assert (
        y == pv("dim1", [StochasticScalar([1, 2, 3]), StochasticScalar([2, 4, 6])])
    ).all()


def test_from_csv():
    x = pv.from_csv("tests/data/variable.csv", "class", "value")
    expected = pv(
        dim_name="class",
        values={
            "Motor": StochasticScalar([0.1, 0.4]),
            "Property": StochasticScalar([0.2, 0.5]),
            "Liability": StochasticScalar([0.3, 0.6]),
        },
    )
    assert (x == expected).all()



from __future__ import annotations
import numpy as np
from numpy.lib.mixins import NDArrayOperatorsMixin
from abc import ABC, abstractmethod
import weakref


class CouplingGroup:
    """A class to represent a group of variables that are coupled together."""

    variables: weakref.WeakSet[ProteusStochasticVariable]

    @property
    def id(self):
        return id(self)

    def __init__(self, variable: ProteusStochasticVariable):
        # Start the group with a single variable, stored as a weak reference.
        self.variables: weakref.WeakSet[ProteusStochasticVariable] = weakref.WeakSet(
            [variable]
        )

    def merge(self, other: CouplingGroup):
        if self is other:
            return
        # Merge the other group's variables into this one, updating their pointer.
        for var in list(other.variables):
            var.coupled_variable_group = self
            self.variables.add(var)
        return


class ProteusStochasticVariable(ABC, NDArrayOperatorsMixin):
    """A class to represent a stochastic variable in a simulation."""

    values: np.ndarray
    n_sims: int

    def __init__(self):
        self.coupled_variable_group = CouplingGroup(self)

    @abstractmethod
    def _reorder_sims(self, new_order: np.ndarray):
        pass

    def all(self):
        return self.values.all()

    def any(self):
        return self.values.any()



from pal import distributions
from pal.config import set_random_seed, xp as np
import pytest
import math
import scipy.special
from scipy.special import gamma


def test_Poisson():
    set_random_seed(12345678910)
    lamda = 3.5
    dist = distributions.Poisson(lamda)
    assert dist.cdf(0) == np.exp(-lamda)
    assert dist.invcdf(0) == 0
    assert np.allclose(
        dist.invcdf(dist.cdf(np.array([0, 2, 5, 10]))),
        np.array([0, 2, 5, 10]),
        1e-8,
    )
    sims = dist.generate(100000)
    assert np.isclose(sims.mean(), lamda, 1e-3)
    assert np.isclose(sims.std() ** 2, lamda, 1e-2)


def test_Poisson_gamma():
    """Tests the Poisson distribution with a gamma distributed lambda."""
    set_random_seed(12345678910)
    alpha = 0.5
    beta = 3
    lamda = distributions.Gamma(alpha, beta).generate(100000)
    sims = distributions.Poisson(lamda).generate(100000)
    sims_mean = sims.mean()
    sims_std = sims.std()
    assert np.isclose(sims_mean, alpha * beta, 1e-2)
    assert np.isclose(sims_std**2, alpha * beta + alpha * beta**2, 1e-2)
    assert sims.coupled_variable_group == lamda.coupled_variable_group


def test_gamma_exp():
    """Tests the Exponential distribution with a gamma distributed lambda."""
    set_random_seed(12345678910)
    alpha = 1.5
    beta = 3
    lamda = distributions.Gamma(alpha, beta).generate(1000000)
    sims = distributions.Exponential(lamda).generate(1000000)
    sims_mean = sims.mean()
    sims_std = sims.std()
    assert np.isclose(sims_mean, alpha * beta, 1e-2)
    assert np.isclose(sims_std**2, (2 * alpha + alpha**2) * beta**2, 1e-2)
    assert sims.coupled_variable_group == lamda.coupled_variable_group


def test_Beta():
    set_random_seed(12345678910)
    alpha = 2
    beta = 3
    scale = 10000000
    loc = 1000000
    dist = distributions.Beta(alpha, beta, scale, loc)
    assert dist.cdf(1000000) == 0.0
    assert dist.invcdf(0) == 1000000
    assert np.allclose(
        dist.invcdf(dist.cdf(np.array([1234560.1, 2345670, 3456780]))),
        np.array([1234560.1, 2345670, 3456780]),
        1e-8,
    )

    sims = dist.generate(1000000)
    assert np.allclose(sims.mean(), alpha / (alpha + beta) * scale + loc, 1e-3)
    assert np.allclose(
        sims.std(),
        math.sqrt(alpha * beta / ((alpha + beta) ** 2 * (alpha + beta + 1))) * scale,
        1e-3,
    )


def test_GPD():
    set_random_seed(12345678910)
    shape = 0.25
    scale = 100000
    threshold = 1000000
    dist = distributions.GPD(shape, scale, threshold)
    assert dist.cdf(1000000) == 0.0
    assert dist.cdf(1500000) == pytest.approx(0.960981557689, 1e-4)
    assert dist.invcdf(0) == 1000000
    assert dist.invcdf(0.960981557689) == pytest.approx(1500000, 1e-4)

    sims = dist.generate(100000000)
    assert sims.mean() == pytest.approx(scale / (1 - shape) + threshold, 1e-3)
    assert sims.std() == pytest.approx(
        scale / (1 - shape) / math.sqrt(1 - 2 * shape), 1e-3
    )


def test_Burr():
    set_random_seed(12345678910)
    power = 2
    shape = 3
    scale = 100000
    loc = 1000000
    dist = distributions.Burr(power, shape, scale, loc)
    assert dist.cdf(1000000) == 0.0
    assert dist.cdf(1500000) == pytest.approx(0.9999431042330451, 1e-8)
    assert dist.invcdf(0) == 1000000
    assert dist.invcdf(0.9999431042330451) == pytest.approx(1500000, 1e-8)

    sims = dist.generate(10000000)
    assert sims.mean() == pytest.approx(
        shape * scipy.special.beta(shape - 1 / power, 1 + 1 / power) * scale + loc, 1e-3
    )
    assert sims.std() == pytest.approx(
        math.sqrt(
            shape * scipy.special.beta(shape - 2 / power, 1 + 2 / power)
            - shape**2 * scipy.special.beta(shape - 1 / power, 1 + 1 / power) ** 2
        )
        * scale,
        1e-3,
    )


def test_InverseBurr():
    set_random_seed(12345678910)
    power = 4
    shape = 5
    scale = 100000
    loc = 1000000
    dist = distributions.InverseBurr(power, shape, scale, loc)
    assert dist.cdf(1000000) == 0.0
    assert dist.invcdf(0) == 1000000

    assert dist.invcdf(0.5) == scale * (1 / (2 ** (1 / shape) - 1)) ** (1 / power) + loc

    sims = dist.generate(10000000)

    assert sims.mean() == pytest.approx(
        gamma(1 - 1 / power) * gamma(shape + 1 / power) / gamma(shape) * scale + loc,
        1e-3,
    )
    assert sims.std() == pytest.approx(
        math.sqrt(
            gamma(1 - 2 / power) * gamma(shape + 2 / power) / gamma(shape)
            - (gamma(1 - 1 / power) * gamma(shape + 1 / power) / gamma(shape)) ** 2
        )
        * scale,
        1e-3,
    )


def test_Logistic():
    set_random_seed(12345678910)
    mu = 2.5
    sigma = 2
    dist = distributions.Logistic(mu, sigma)
    assert dist.cdf(2.5) == 0.5
    assert dist.invcdf(0.5) == 2.5
    assert np.allclose(
        dist.invcdf(dist.cdf(np.array([1.1, 2, 3]))),
        np.array([1.1, 2, 3]),
    )

    sims = dist.generate(10000000)
    sims_mean = sims.mean()
    sims_std = sims.std()
    assert np.isclose(sims_mean, mu, 1e-3)
    assert np.isclose(sims_std, np.pi * sigma / np.sqrt(3), 1e-3)


def test_LogLogistic():
    set_random_seed(12345678910)
    shape = 4
    scale = 100000
    loc = 1000000
    dist = distributions.LogLogistic(shape, scale, loc)
    assert dist.cdf(1000000) == 0.0
    assert dist.cdf(1500000) == pytest.approx(0.9984025559105432, 1e-8)
    assert dist.invcdf(0) == 1000000
    assert dist.invcdf(0.5) == scale + loc
    assert dist.invcdf(0.9984025559105432) == pytest.approx(1500000, 1e-8)

    sims = dist.generate(100000000)
    assert sims.mean() == pytest.approx(
        scipy.special.beta(1 - 1 / shape, 1 + 1 / shape) * scale + loc, 1e-3
    )
    assert sims.std() == pytest.approx(
        math.sqrt(
            scipy.special.beta(1 - 2 / shape, 1 + 2 / shape)
            - scipy.special.beta(1 - 1 / shape, 1 + 1 / shape) ** 2
        )
        * scale,
        1e-3,
    )


def test_ParaLogistic():
    set_random_seed(12345678910)
    shape = 2.5
    scale = 100000
    loc = 1000000
    dist = distributions.Paralogistic(shape, scale, loc)

    assert dist.cdf(1000000) == 0.0
    assert dist.invcdf(0) == 1000000
    assert dist.invcdf(
        dist.cdf(np.array([1234560.1, 2345670, 3456780]))
    ) == pytest.approx(np.array([1234560.1, 2345670, 3456780]), 1e-8)

    sims = dist.generate(100000000)

    assert sims.mean() == pytest.approx(
        scale * gamma(1 + 1 / shape) * gamma(shape - 1 / shape) / gamma(shape) + loc,
        1e-5,
    )


def test_InverseParaLogistic():
    set_random_seed(12345678910)
    shape = 5
    scale = 100000
    loc = 1000000
    dist = distributions.InverseParalogistic(shape, scale, loc)

    assert dist.cdf(1000000) == 0.0
    assert dist.invcdf(0) == 1000000
    assert dist.invcdf(
        dist.cdf(np.array([1234560.1, 2345670, 3456780]))
    ) == pytest.approx(np.array([1234560.1, 2345670, 3456780]), 1e-8)

    sims = dist.generate(100000000)

    assert sims.mean() == pytest.approx(
        scale * gamma(shape + 1 / shape) * gamma(1 - 1 / shape) / gamma(shape) + loc,
        1e-3,
    )
    assert sims.std() == pytest.approx(
        scale
        * np.sqrt(
            (gamma(shape + 2 / shape) * gamma(1 - 2 / shape) / gamma(shape))
            - (gamma(shape + 1 / shape) * gamma(1 - 1 / shape) / gamma(shape)) ** 2
        ),
        1e-3,
    )


def test_Weibull():
    set_random_seed(12345678910)
    shape = 2
    scale = 1000000
    loc = 1000000
    dist = distributions.Weibull(shape, scale, loc)

    assert dist.cdf(1000000) == 0.0
    assert dist.invcdf(0) == 1000000
    assert dist.invcdf(
        dist.cdf(np.array([1234560.1, 2345670, 3456780]))
    ) == pytest.approx(np.array([1234560.1, 2345670, 3456780]), 1e-8)

    sims = dist.generate(100000000)

    assert sims.mean() == pytest.approx(scale * gamma(1 + 1 / shape) + loc, 1e-3)
    assert sims.std() == pytest.approx(
        scale * np.sqrt(gamma(1 + 2 / shape) - (gamma(1 + 1 / shape)) ** 2), 1e-3
    )


def test_InverseWeibull():
    set_random_seed(12345678910)
    shape = 4
    scale = 1000000
    loc = 1000000
    dist = distributions.InverseWeibull(shape, scale, loc)

    assert dist.cdf(1000000) == 0.0
    assert dist.invcdf(0) == 1000000
    assert dist.invcdf(
        dist.cdf(np.array([1234560.1, 2345670, 3456780]))
    ) == pytest.approx(np.array([1234560.1, 2345670, 3456780]), 1e-8)

    sims = dist.generate(100000000)

    assert sims.mean() == pytest.approx(scale * gamma(1 - 1 / shape) + loc, 1e-3)
    assert sims.std() == pytest.approx(
        scale * np.sqrt(gamma(1 - 2 / shape) - (gamma(1 - 1 / shape)) ** 2), 1e-3
    )


def test_Exponential():
    set_random_seed(12345678910)
    scale = 1000000
    loc = 1000000
    dist = distributions.Exponential(scale, loc)

    assert dist.cdf(1000000) == 0.0
    assert dist.invcdf(0) == 1000000
    assert dist.invcdf(
        dist.cdf(np.array([1234560.1, 2345670, 3456780]))
    ) == pytest.approx(np.array([1234560.1, 2345670, 3456780]), 1e-8)

    sims = dist.generate(100000000)

    assert sims.mean() == pytest.approx(scale + loc, 1e-3)
    assert sims.std() == pytest.approx(scale, 1e-3)


def test_InverseExponential():
    set_random_seed(12345678910)
    scale = 1000000
    loc = 1000000
    dist = distributions.InverseExponential(scale, loc)

    assert dist.cdf(1000000) == 0.0
    assert dist.invcdf(0) == 1000000
    assert dist.invcdf(
        dist.cdf(np.array([1234560.1, 2345670, 3456780]))
    ) == pytest.approx(np.array([1234560.1, 2345670, 3456780]), 1e-8)


def test_Gamma():
    set_random_seed(12345678910)
    scale = 1000000
    shape = 4.5
    loc = 1000000
    dist = distributions.Gamma(shape, scale, loc)

    assert dist.cdf(1000000) == 0.0
    assert dist.invcdf(0) == 1000000
    assert np.allclose(
        dist.invcdf(dist.cdf(np.array([1234560.1, 2345670, 3456780]))),
        np.array([1234560.1, 2345670, 3456780]),
        1e-8,
    )

    sims = dist.generate(10000000)

    assert np.allclose(sims.mean(), scale * shape + loc, 1e-3)
    assert np.allclose(sims.std(), scale * np.sqrt(shape), 1e-3)


def test_LogNormal():
    set_random_seed(12345678910)
    mu = 8
    sigma = 1.25
    dist = distributions.LogNormal(mu, sigma)

    assert dist.cdf(0) == 0.0
    assert dist.invcdf(0) == 0
    assert np.allclose(
        dist.invcdf(dist.cdf(np.array([1234560.1, 2345670, 3456780]))),
        np.array([1234560.1, 2345670, 3456780]),
        1e-8,
    )

    sims = dist.generate(100000000)

    mean = np.exp(mu + 0.5 * sigma**2)
    sd = np.sqrt((np.exp(sigma**2) - 1) * np.exp(2 * mu + sigma**2))

    assert np.allclose(sims.mean(), mean, 1e-3)
    assert np.allclose(sims.std(), sd, 1e-3)


def test_InverseGamma():
    set_random_seed(12345678910)
    scale = 1000000
    shape = 3.5
    loc = 1000000
    dist = distributions.InverseGamma(shape, scale, loc)

    assert dist.cdf(1000000) == 0.0
    assert dist.invcdf(0) == 1000000
    assert np.allclose(
        dist.invcdf(dist.cdf(np.array([1234560.1, 2345670, 3456780]))),
        np.array([1234560.1, 2345670, 3456780]),
        1e-8,
    )

    sims = dist.generate(10000000)

    assert np.allclose(sims.mean(), scale * gamma(shape - 1) / gamma(shape) + loc, 1e-3)
    assert np.allclose(
        sims.std(),
        scale
        * np.sqrt(
            gamma(shape - 2) / gamma(shape) - (gamma(shape - 1) / gamma(shape)) ** 2
        ),
        1e-3,
    )



from pal import copulas
from pal.variables import StochasticScalar
import pytest  # noqa


def test_copula_reordering():
    """A check that the copula reordering works as expected."""
    x = StochasticScalar([4, 5, 2, 1, 3])
    y = StochasticScalar([1, 2, 3, 4, 5])
    copula_samples = [
        StochasticScalar([0, 4, 3, 1, 2]),
        StochasticScalar([1, 3, 2, 0, 4]),
    ]
    copulas.apply_copula([x, y], copula_samples)
    assert (x.values == [4, 5, 2, 1, 3]).all()
    assert (y.values == [3, 4, 1, 2, 5]).all()


def test_coupled_variable_reordering():
    """Test that coupled variables are reordered correctly."""
    x = StochasticScalar([4, 5, 2, 1, 3])
    y = StochasticScalar([1, 2, 3, 4, 5])
    z = y + 1  # y and z are now coupled
    copula_samples = [
        StochasticScalar([0, 4, 3, 1, 2]),
        StochasticScalar([1, 3, 2, 0, 4]),
    ]
    copulas.apply_copula([x, y], copula_samples)
    assert (x.values == [4, 5, 2, 1, 3]).all()
    assert (y.values == [3, 4, 1, 2, 5]).all()
    assert (z.values == [4, 5, 2, 3, 6]).all()
    assert (
        x.coupled_variable_group == y.coupled_variable_group == z.coupled_variable_group
    )


def test_coupled_variable_reordering2():
    """Test that coupled variables are reordered correctly."""
    x = StochasticScalar([4, 5, 2, 1, 3])
    y = StochasticScalar([1, 2, 3, 4, 5])
    z = StochasticScalar([7, 3, 1, 9, 0])
    a = y + z  # a, y, and z are now coupled
    copula_samples = [
        StochasticScalar([0, 4, 3, 1, 2]),
        StochasticScalar([1, 3, 2, 0, 4]),
    ]
    copulas.apply_copula([x, y], copula_samples)
    assert (x.values == [4, 5, 2, 1, 3]).all()
    assert (y.values == [3, 4, 1, 2, 5]).all()
    assert (z.values == [1, 9, 7, 3, 0]).all()
    assert (a.values == [4, 13, 8, 5, 5]).all()
    assert (
        x.coupled_variable_group
        == y.coupled_variable_group
        == z.coupled_variable_group
        == a.coupled_variable_group
    )



from __future__ import annotations
from typing import Union, overload
from .frequency_severity import FreqSevSims
from .stochastic_scalar import StochasticScalar, ProteusStochasticVariable
import numpy as np
import scipy.stats  # type: ignore
import pandas as pd  # type: ignore
import plotly.graph_objects as go  # type: ignore
import plotly.io as pio  # type: ignore

pio.templates.default = "none"


class ProteusVariable:
    """A class to hold a multivariate variable in a simulation.

    A Proteus Variable is a hierarchical structure that can hold multiple
    scalar variables. The purpose of this class is to allow
    for the creation of more complex variables that can be used in
    simulations.

    Each level of a Proteus Variable can be a list or dictionary of scalar variables or other ProteusVariable objects. Each level can have a different number of elements.
    Each level has a name that can be used to access the level in the hierarchy.

    Sub elements of a ProteusVariable can be accessed using the [] notation.

    """

    def __len__(self):
        return len(self.values)

    def __init__(
        self,
        dim_name: str,
        values: (
            list[Union[ProteusVariable, StochasticScalar | FreqSevSims | float | int]]
            | dict[
                str,
                Union[ProteusVariable, StochasticScalar | FreqSevSims | float | int],
            ]
        ),
    ):
        self.dim_name: str = dim_name
        self.values = values
        self.dimensions = [dim_name]
        self._dimension_set = set(self.dimensions)
        # check the number of simulations in each variable
        self.n_sims = None
        for value in (
            self.values.values() if isinstance(self.values, dict) else self.values
        ):
            if isinstance(value, ProteusVariable):
                if (
                    self._dimension_set.intersection(value._dimension_set)
                    or self.dim_name == value.dim_name
                ):
                    raise ValueError(
                        "Duplicate dimension names in ProteusVariable hierarchy."
                    )
                self._dimension_set.intersection_update(value.dimensions)
                self.dimensions.extend(value.dimensions)

            if self.n_sims is None:
                if isinstance(value, ProteusStochasticVariable):
                    self.n_sims = value.n_sims
                else:
                    self.n_sims = 1
            elif isinstance(value, ProteusStochasticVariable):
                if value.n_sims != self.n_sims:
                    if self.n_sims == 1:
                        self.n_sims == value.n_sims
                    else:
                        raise ValueError("Number of simulations do not match.")

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        if method != "__call__":
            return NotImplemented

        def recursive_apply(*items):
            # If none of the items is a ProteusVariable (i.e. a container), then
            # assume they are leaf nodes (e.g., numbers or stochastic types) and simply call ufunc.
            if not any(isinstance(item, ProteusVariable) for item in items):
                # For stochastic types that implement __array_ufunc__, this call will
                # automatically delegate to their own __array_ufunc__.
                return ufunc(*items, **kwargs)

            # Otherwise, at least one of the items is a container.
            # We assume that the container structure is consistent across items.
            first = items[0]
            if isinstance(first, ProteusVariable):
                # Process dictionary containers.
                if isinstance(first.values, dict):
                    new_data = {}
                    # Iterate over each key in the container.
                    for key in first.values:
                        new_items = []
                        for item in items:
                            if isinstance(item, ProteusVariable):
                                new_items.append(item.values[key])
                            else:
                                new_items.append(item)
                        new_data[key] = recursive_apply(*new_items)
                    return ProteusVariable(first.dim_name, new_data)
                # Process list containers.
                elif isinstance(first.values, list):
                    new_list = []
                    for idx, _ in enumerate(first.values):
                        new_items = []
                        for item in items:
                            if isinstance(item, ProteusVariable):
                                new_items.append(item.values[idx])
                            else:
                                new_items.append(item)
                        new_list.append(recursive_apply(*new_items))
                    return ProteusVariable(first.dim_name, new_list)
                else:
                    # In case data is neither dict nor list, try applying ufunc directly.
                    return ufunc(first.values, **kwargs)
            else:
                # If the first item is not a container but some other item is,
                # we assume they can all be processed by ufunc.
                return ufunc(*items, **kwargs)

        return recursive_apply(*inputs)

    def __array_function__(self, func, types, args, kwargs):
        args = [
            (
                (
                    list(arg.values.values())
                    if isinstance(arg.values, dict)
                    else arg.values
                )
                if isinstance(arg, ProteusVariable)
                else arg
            )
            for arg in args
        ]
        temp = func(*args, **kwargs)
        if isinstance(self.values, dict):
            return ProteusVariable(
                self.dim_name,
                {key: temp[i] for i, key in enumerate(self.values.keys())},
            )
        else:
            return ProteusVariable(self.dim_name, [value for value in temp])

    @overload
    def sum(self) -> StochasticScalar: ...
    @overload
    def sum(self, dimensions: list[str]) -> ProteusVariable: ...

    def sum(
        self, dimensions: list[str] = []
    ) -> Union[ProteusVariable, StochasticScalar]:
        """Sum the variables across the specified dimensions. Returns a new ProteusVariable with the summed values."""
        if dimensions is None or dimensions == []:
            result: StochasticScalar = sum(self)
            return result
        if self.dimensions in dimensions:
            result = ProteusVariable(dim_name=self.values[0].dimensions, values=0)
            for value in self.values:
                if isinstance(value, ProteusVariable | StochasticScalar):
                    result = result + value.sum(dimensions)
                else:
                    result = result + value
            return result
        else:
            return self

    def __iter__(self):
        if isinstance(self.values, dict):
            return iter(self.values.values())
        else:
            return iter(self.values)

    def _binary_operation(self, other, operation):
        if isinstance(other, ProteusVariable):
            if self.dimensions != other.dimensions:
                raise ValueError("Dimensions of the two variables do not match.")
        if isinstance(self.values, dict):
            if isinstance(other, ProteusVariable):
                return ProteusVariable(
                    dim_name=self.dim_name,
                    values={
                        key: operation(value, other.values[key])
                        for key, value in self.values.items()
                    },
                )
            return ProteusVariable(
                dim_name=self.dim_name,
                values={
                    key: operation(value, other) for key, value in self.values.items()
                },
            )
        elif isinstance(self.values, list):
            if isinstance(other, ProteusVariable):
                return ProteusVariable(
                    dim_name=self.dim_name,
                    values=[
                        operation(value, other.values[i])
                        for i, value in enumerate(self.values)
                    ],
                )
            return ProteusVariable(
                dim_name=self.dim_name,
                values=[operation(value, other) for i, value in enumerate(self.values)],
            )

    def __add__(self, other) -> ProteusVariable:
        return self._binary_operation(other, lambda a, b: a + b)

    def __radd__(self, other) -> ProteusVariable:
        return self.__add__(other)

    def __mul__(self, other) -> ProteusVariable:
        return self._binary_operation(other, lambda a, b: a * b)

    def __rmul__(self, other) -> ProteusVariable:
        return self.__mul__(other)

    def __sub__(self, other) -> ProteusVariable:
        return self._binary_operation(other, lambda a, b: a - b)

    def __rsub__(self, other) -> ProteusVariable:
        return self._binary_operation(other, lambda a, b: b - a)

    def __ge__(self, other) -> ProteusVariable:
        return self._binary_operation(other, lambda a, b: a >= b)

    def __le__(self, other) -> ProteusVariable:
        return self._binary_operation(other, lambda a, b: a <= b)

    def __gt__(self, other) -> ProteusVariable:
        return self._binary_operation(other, lambda a, b: a > b)

    def __lt__(self, other) -> ProteusVariable:
        return self._binary_operation(other, lambda a, b: a < b)

    def __rge__(self, other) -> ProteusVariable:
        return self.__lt__(other)

    def __rle__(self, other) -> ProteusVariable:
        return self.__gt__(other)

    def __rgt__(self, other) -> ProteusVariable:
        return self.__le__(other)

    def __rlt__(self, other) -> ProteusVariable:
        return self.__ge__(other)

    def __getitem__(self, key: str | int):
        if isinstance(self.values, dict):
            if isinstance(key, int):
                return self.values[list(self.values.keys())[key]]
            else:
                return self.values[key]
        else:
            if isinstance(key, int):
                return self.values[key]
            else:
                raise ValueError("Key must be an integer for a list.")

    def get_value_at_sim(self, sim_no: int | StochasticScalar):
        _get_value = lambda x: (  # noqa : E731
            x.get_value_at_sim(sim_no)
            if isinstance(x, ProteusVariable)
            else x[sim_no] if x.n_sims > 1 else x
        )
        if isinstance(self.values, dict):
            result = ProteusVariable(
                dim_name=self.dim_name,
                values={k: _get_value(v) for k, v in self.values.items()},
            )
        elif isinstance(self.values, list):
            result = ProteusVariable(
                dim_name=self.dim_name,
                values=[_get_value(v) for v in self.values],
            )
        return result

    def all(self) -> bool:
        if isinstance(self.values, dict):
            return all([value.all() for value in self.values.values()])
        else:
            return all([value.all() for value in self.values])

    def any(self) -> bool:
        if isinstance(self.values, dict):
            return any([value.any() for value in self.values.values()])
        else:
            return any([value.any() for value in self.values])

    def upsample(self, n_sims: int) -> ProteusVariable:
        """Upsample the variable to the specified number of simulations"""
        if self.n_sims == n_sims:
            return self
        if isinstance(self.values, dict):
            return ProteusVariable(
                dim_name=self.dim_name,
                values={
                    key: (
                        value.upsample(n_sims)
                        if isinstance(value, ProteusStochasticVariable)
                        else value
                    )
                    for key, value in self.values.items()
                },
            )
        else:
            return ProteusVariable(
                dim_name=self.dim_name,
                values=[
                    (
                        value.upsample(n_sims)
                        if isinstance(value, ProteusStochasticVariable)
                        else value
                    )
                    for value in self.values
                ],
            )

    def __eq__(self, other):
        return self._binary_operation(other, lambda a, b: a == b)

    @classmethod
    def from_csv(
        cls,
        file_name: str,
        dim_name: str,
        values_column: str,
        simulation_column="Simulation",
    ) -> ProteusVariable:
        """Import a ProteusVariable from a CSV file.

        Note that only one dimensional variables are supported.
        """
        df = pd.read_csv(file_name)
        pivoted_df = df.pivot(
            index=simulation_column, columns=dim_name, values=values_column
        )
        count = df[dim_name].value_counts()
        pivoted_df.sort_index(inplace=True)

        result = cls(
            dim_name,
            {
                str(label): StochasticScalar(pivoted_df[label].values[: count[label]])
                for label in df[dim_name].unique()
            },
        )
        result.n_sims = max(count)

        return result

    @classmethod
    def from_dict(cls, data: dict[str, list[float]]) -> ProteusVariable:
        """Import a ProteusVariable from a dictionary.

        Note that only one dimensional variables are supported.
        """
        result = cls(
            dim_name="Dim1",
            values={str(label): StochasticScalar(data[label]) for label in data.keys()},
        )
        result.n_sims = max([len(v) for v in data.values()])

        return result

    @classmethod
    def from_series(cls, data: pd.Series) -> ProteusVariable:
        """Create a ProteusVariable from a pandas Series.

        Note that only one dimensional variables are supported.
        """
        result = cls(
            dim_name=data.index.name,
            values={label: data[label] for label in data.index},
        )
        result.n_sims = 1

        return result

    def __repr__(self):
        return f"ProteusVariable(dim_name={self.dim_name}, values={self.values})"

    def correlation_matrix(self, correlation_type="spearman") -> list[list[float]]:
        # validate type
        correlation_type = correlation_type.lower()
        assert correlation_type in ["linear", "spearman", "kendall"]
        assert hasattr(self[0], "values")
        n = len(self.values)
        result: list[list[float]] = [[0.0] * n] * n
        values = [self[i] for i in range(len(self.values))]
        if correlation_type.lower() in ["spearman", "kendall"]:
            # rank the variables first
            for i, value in enumerate(values):
                values[i] = scipy.stats.rankdata(value.values)

        if correlation_type == "kendall":
            for i, value1 in enumerate(values):
                for j, value2 in enumerate(values):
                    result[i][j] = scipy.stats.kendalltau(value1, value2)
        else:
            result = np.corrcoef(values).tolist()

        return result

    def show_histogram(self, title: str | None = None):
        """Show a histogram of the variable values.

        Args:
            title (str | None): The title of the histogram. If None, no title is set.

        """
        fig = go.Figure(layout=go.Layout(title=title))
        labels = (
            self.values.keys()
            if isinstance(self.values, dict)
            else range(len(self.values))
        )
        for value, label in zip(self.values.values(), labels):
            fig.add_trace(go.Histogram(x=value.values, name=label))
        fig.show()

    def show_cdf(self, title: str | None = None):
        """Show a plot of the cumulative distribution function (cdf) of the variable values.

        Args:
            title (str | None): The title of the cdf. If None, no title is set.

        """

        fig = go.Figure(layout=go.Layout(title=title))
        labels = (
            self.values.keys()
            if isinstance(self.values, dict)
            else range(len(self.values))
        )
        for value, label in zip(self.values.values(), labels):
            fig.add_trace(
                go.Scatter(
                    x=np.sort(value.values),
                    y=np.arange(value.n_sims) / value.n_sims,
                    name=label,
                )
            )
        fig.update_xaxes(title_text="Value")
        fig.update_yaxes(title_text="Cumulative Probability")
        fig.show()



"""
Distributions Module

This module contains classes for simulating statistical distributions.
The implementations follow conventions similar to Klugman. Random number
generation and GPU support are managed via configuration settings.
"""

# Standard library imports
from abc import ABC, abstractmethod
from typing import Any, Dict, Union, TypeVar

# Local imports
from .config import config, xp as np, _use_gpu
from .stochastic_scalar import StochasticScalar

# Use cupyx if GPU is enabled
if _use_gpu:
    import cupyx.scipy.special as special  # type: ignore
else:
    import scipy.special as special  # type: ignore

# Type aliases
NumberType = Union[float, int]
NumberOrStochasticScalar = TypeVar(
    "NumberOrStochasticScalar", NumberType, StochasticScalar
)


class DistributionBase(ABC):
    """Abstract base class for statistical distributions."""

    def __init__(self, **params: Any) -> None:
        # Store parameters in a private dictionary.
        self._params: Dict[str, Any] = params

    @property
    def _param_values(self):
        # Yields parameter values; if a parameter is a StochasticScalar, its 'values' are returned.
        for param in self._params.values():
            yield param.values if isinstance(param, StochasticScalar) else param

    @abstractmethod
    def cdf(self, x: NumberOrStochasticScalar) -> NumberOrStochasticScalar:
        """Compute the cumulative distribution function at x."""
        pass

    @abstractmethod
    def invcdf(self, u: NumberOrStochasticScalar) -> NumberOrStochasticScalar:
        """Compute the inverse cumulative distribution function at u."""
        pass

    def generate(
        self, n_sims: int | None = None, rng: np.random.Generator = config.rng
    ) -> StochasticScalar:
        """
        Generate random samples from the distribution.

        Parameters:
            n_sims (int, optional): Number of simulations. Uses config.n_sims if None.
            rng (np.random.Generator, optional): Random number generator.

        Returns:
            StochasticScalar: Generated samples.
        """
        if n_sims is None:
            n_sims = config.n_sims

        result = self._generate(n_sims, rng)
        # Merge coupled variable groups from parameters if applicable.
        for param in self._params.values():
            if isinstance(param, StochasticScalar):
                result.coupled_variable_group.merge(param.coupled_variable_group)
        return result

    def _generate(self, n_sims: int, rng: np.random.Generator) -> StochasticScalar:
        # Default generation method using the inverse CDF technique.
        return StochasticScalar(self.invcdf(rng.uniform(size=n_sims)))


class DiscreteDistributionBase(DistributionBase, ABC):
    """Abstract base class for discrete distributions."""

    # Inherits __init__, cdf, and invcdf abstract methods from DistributionBase.
    pass


# --- Discrete Distributions ---


class Poisson(DiscreteDistributionBase):
    """Poisson Distribution.

    Parameters:
        mean (float): Mean number of events.
    """

    def __init__(self, mean: float) -> None:
        super().__init__(mean=mean)

    def cdf(self, x: NumberOrStochasticScalar) -> NumberOrStochasticScalar:
        return special.pdtr(x, self._params["mean"])

    def invcdf(self, u: NumberOrStochasticScalar) -> NumberOrStochasticScalar:
        return special.pdtrik(u, self._params["mean"])

    def _generate(self, n_sims: int, rng: np.random.Generator) -> StochasticScalar:
        (mean,) = self._param_values
        return StochasticScalar(rng.poisson(mean, n_sims))


class NegBinomial(DiscreteDistributionBase):
    """Negative Binomial Distribution.

    Parameters:
        n (float or StochasticScalar): Number of failures until stop.
        p (float or StochasticScalar): Probability of success.
    """

    def __init__(
        self, n: Union[float, StochasticScalar], p: Union[float, StochasticScalar]
    ) -> None:
        super().__init__(n=n, p=p)

    def cdf(self, x: NumberOrStochasticScalar) -> NumberOrStochasticScalar:
        n, p = self._param_values
        return special.nbdtr(x, n, p)

    def invcdf(self, u: NumberOrStochasticScalar) -> NumberOrStochasticScalar:
        n, p = self._param_values
        return special.nbdtri(u, n, p)

    def _generate(self, n_sims: int, rng: np.random.Generator) -> StochasticScalar:
        n, p = self._param_values
        return StochasticScalar(rng.negative_binomial(n, p, n_sims))


class Binomial(DiscreteDistributionBase):
    """Binomial Distribution.

    Parameters:
        n (int): Number of trials.
        p (float): Probability of success.
    """

    def __init__(self, n: int, p: float) -> None:
        super().__init__(n=n, p=p)

    def cdf(self, x: NumberOrStochasticScalar) -> NumberOrStochasticScalar:
        n, p = self._param_values
        return special.bdtr(x, n, p)

    def invcdf(self, u: NumberOrStochasticScalar) -> NumberOrStochasticScalar:
        n, p = self._param_values
        return special.bdtri(u, n, p)

    def _generate(self, n_sims: int, rng: np.random.Generator) -> StochasticScalar:
        n, p = self._param_values
        return StochasticScalar(rng.binomial(n, p, n_sims))


class HyperGeometric(DiscreteDistributionBase):
    """Hypergeometric Distribution.

    Models the number of successes in draws without replacement.

    Parameters:
        ngood (int): Number of good items.
        nbad (int): Number of bad items.
        population_size (int): Total population size.
    """

    def __init__(self, ngood: int, nbad: int, population_size: int) -> None:
        # Note: population_size is stored with key 'n'
        super().__init__(ngood=ngood, nbad=nbad, n=population_size)

    def cdf(self, x: NumberOrStochasticScalar) -> NumberOrStochasticScalar:
        raise NotImplementedError("CDF for HyperGeometric is not implemented.")

    def invcdf(self, u: NumberOrStochasticScalar) -> NumberOrStochasticScalar:
        raise NotImplementedError("Inverse CDF for HyperGeometric is not implemented.")

    def _generate(self, n_sims: int, rng: np.random.Generator) -> StochasticScalar:
        ngood, nbad, n = self._param_values
        return StochasticScalar(rng.hypergeometric(ngood, nbad, n, n_sims))


# --- Continuous Distributions ---


class GPD(DistributionBase):
    r"""Generalized Pareto Distribution.

    Defined by:
        F(x) = 1 - (1 + ξ(x-μ)/σ)^(-1/ξ) for ξ ≠ 0,
        F(x) = 1 - exp(-(x-μ)/σ) for ξ = 0.

    Parameters:
        shape: ξ, the shape parameter.
        scale: σ, the scale parameter.
        loc: μ, the location parameter.
    """

    def __init__(
        self,
        shape: NumberOrStochasticScalar,
        scale: NumberOrStochasticScalar,
        loc: NumberOrStochasticScalar,
    ) -> None:
        super().__init__(shape=shape, scale=scale, loc=loc)

    def cdf(self, x: NumberOrStochasticScalar) -> NumberOrStochasticScalar:
        shape, scale, loc = self._params.values()
        if shape != 0:
            return 1 - (1 + shape * (x - loc) / scale) ** (-1 / shape)
        else:
            return 1 - np.exp(-(x - loc) / scale)

    def invcdf(self, u: NumberOrStochasticScalar) -> NumberOrStochasticScalar:
        shape, scale, loc = self._params.values()
        return (np.exp(np.log(1 - u) * (-shape)) - 1) * (scale / shape) + loc


class Burr(DistributionBase):
    r"""Burr Distribution.

    Defined by:
        F(x) = 1 - [1 + ((x-μ)/σ)^power]^(-shape), x > μ

    Parameters:
        power: The power parameter.
        shape: The shape parameter.
        scale: The scale parameter.
        loc: The location parameter.
    """

    def __init__(
        self,
        power: NumberOrStochasticScalar,
        shape: NumberOrStochasticScalar,
        scale: NumberOrStochasticScalar,
        loc: NumberOrStochasticScalar,
    ) -> None:
        super().__init__(power=power, shape=shape, scale=scale, loc=loc)

    def cdf(self, x: NumberOrStochasticScalar) -> NumberOrStochasticScalar:
        power, shape, scale, loc = self._params.values()
        return 1 - (1 + ((x - loc) / scale) ** power) ** (-shape)

    def invcdf(self, u: NumberOrStochasticScalar) -> NumberOrStochasticScalar:
        power, shape, scale, loc = self._params.values()
        return scale * (((1 / (1 - u)) ** (1 / shape) - 1) ** (1 / power)) + loc


class Beta(DistributionBase):
    r"""Beta Distribution.

    Defined by:
        F(x) = (Γ(α+β) / (Γ(α)Γ(β))) ∫₀^((x-μ)/σ) u^(α-1)(1-u)^(β-1) du

    Parameters:
        alpha (float): Alpha parameter.
        beta (float): Beta parameter.
        scale (float): Scale parameter (default 1).
        loc (float): Location parameter (default 0).
    """

    def __init__(
        self, alpha: float, beta: float, scale: float = 1, loc: float = 0
    ) -> None:
        super().__init__(alpha=alpha, beta=beta, scale=scale, loc=loc)

    def cdf(self, x: NumberOrStochasticScalar) -> NumberOrStochasticScalar:
        alpha, beta, scale, loc = self._params.values()
        return special.betainc(alpha, beta, (x - loc) / scale)

    def invcdf(self, u: NumberOrStochasticScalar) -> NumberOrStochasticScalar:
        alpha, beta, scale, loc = self._params.values()
        return special.betaincinv(alpha, beta, u) * scale + loc


class LogLogistic(DistributionBase):
    r"""Log-Logistic Distribution.

    Defined by:
        F(x) = y / (1 + y) where y = ((x-μ)/σ)^shape, x > μ

    Parameters:
        shape (float): Shape parameter.
        scale (float): Scale parameter.
        loc (float): Location parameter (default 0).
    """

    def __init__(self, shape: float, scale: float, loc: float = 0) -> None:
        super().__init__(shape=shape, scale=scale, loc=loc)

    def cdf(self, x: NumberOrStochasticScalar) -> NumberOrStochasticScalar:
        shape, scale, loc = self._params.values()
        y = ((x - loc) / scale) ** shape
        return y / (1 + y)

    def invcdf(self, u: NumberOrStochasticScalar) -> NumberOrStochasticScalar:
        shape, scale, loc = self._params.values()
        return scale * ((u / (1 - u)) ** (1 / shape)) + loc


class Normal(DistributionBase):
    """Normal Distribution.

    Parameters:
        mu (float): Mean.
        sigma (float): Standard deviation.
    """

    def __init__(self, mu: float, sigma: float) -> None:
        super().__init__(mu=mu, sigma=sigma)

    def cdf(self, x: NumberOrStochasticScalar) -> NumberOrStochasticScalar:
        mu, sigma = self._param_values
        return special.ndtr((x - mu) / sigma)

    def invcdf(self, u: NumberOrStochasticScalar) -> NumberOrStochasticScalar:
        mu, sigma = self._param_values
        return special.ndtri(u) * sigma + mu


class Logistic(DistributionBase):
    """Logistic Distribution.

    Parameters:
        mu (float): Location parameter.
        sigma (float): Scale parameter.
    """

    def __init__(self, mu: float, sigma: float) -> None:
        super().__init__(mu=mu, sigma=sigma)

    def cdf(self, x: NumberOrStochasticScalar) -> NumberOrStochasticScalar:
        mu, sigma = self._param_values
        return 1 / (1 + np.exp(-(x - mu) / sigma))

    def invcdf(self, u: NumberOrStochasticScalar) -> NumberOrStochasticScalar:
        mu, sigma = self._param_values
        return mu + sigma * np.log(u / (1 - u))


class LogNormal(DistributionBase):
    """Log-Normal Distribution.

    Parameters:
        mu (float): Mean of the logged variable.
        sigma (float): Standard deviation of the logged variable.
    """

    def __init__(self, mu: float, sigma: float) -> None:
        super().__init__(mu=mu, sigma=sigma)

    def cdf(self, x: NumberOrStochasticScalar) -> NumberOrStochasticScalar:
        mu, sigma = self._param_values
        return special.ndtr((np.log(x) - mu) / sigma)

    def invcdf(self, u: NumberOrStochasticScalar) -> NumberOrStochasticScalar:
        mu, sigma = self._param_values
        return np.exp(special.ndtri(u) * sigma + mu)


class Gamma(DistributionBase):
    r"""Gamma Distribution.

    Defined by:
        F(x) = (1/Γ(α)) γ(α, (x-μ)/θ), x > μ

    Parameters:
        alpha: Shape parameter.
        theta: Scale parameter.
        loc: Location parameter (default 0).
    """

    def __init__(self, alpha: float, theta: float, loc: float = 0) -> None:
        super().__init__(alpha=alpha, theta=theta, loc=loc)

    def cdf(self, x: NumberOrStochasticScalar) -> NumberOrStochasticScalar:
        alpha, theta, loc = self._param_values
        return special.gammainc(alpha, (x - loc) / theta)

    def invcdf(self, u: NumberOrStochasticScalar) -> NumberOrStochasticScalar:
        alpha, theta, loc = self._param_values
        return special.gammaincinv(alpha, u) * theta + loc

    def _generate(self, n_sims: int, rng: np.random.Generator) -> StochasticScalar:
        alpha, theta, loc = self._param_values
        return StochasticScalar(rng.gamma(alpha, theta, size=n_sims) + loc)


class InverseGamma(DistributionBase):
    r"""Inverse Gamma Distribution.

    Defined by:
        F(x) = 1 - (1/Γ(α)) γ(α, θ/(x-μ)), x > μ

    Parameters:
        alpha: Shape parameter.
        theta: Scale parameter.
        loc: Location parameter (default 0).
    """

    def __init__(self, alpha: float, theta: float, loc: float = 0) -> None:
        super().__init__(alpha=alpha, theta=theta, loc=loc)

    def cdf(self, x: NumberOrStochasticScalar) -> NumberOrStochasticScalar:
        alpha, theta, loc = self._param_values
        return special.gammaincc(alpha, np.divide(theta, (x - loc)))

    def invcdf(self, u: NumberOrStochasticScalar) -> NumberOrStochasticScalar:
        alpha, theta, loc = self._param_values
        return np.divide(theta, special.gammainccinv(alpha, u)) + loc


class Pareto(DistributionBase):
    r"""Pareto Distribution.

    Defined by:
        F(x) = 1 - (x_m / x)^a

    Parameters:
        shape (float): Shape parameter.
        scale (float): Scale parameter (x_m).
    """

    def __init__(self, shape: float, scale: float) -> None:
        super().__init__(shape=shape, scale=scale)

    def cdf(self, x: NumberOrStochasticScalar) -> NumberOrStochasticScalar:
        shape, scale = self._param_values
        return 1 - (x / scale) ** (-shape)

    def invcdf(self, u: NumberOrStochasticScalar) -> NumberOrStochasticScalar:
        shape, scale = self._param_values
        return (1 - u) ** (-1 / shape) * scale


class Paralogistic(DistributionBase):
    r"""ParaLogistic Distribution.

    Defined by:
        F(x) = 1 - [1 + ((x-μ)/σ)^α]^(-α), x > μ

    Parameters:
        shape (float): Shape parameter.
        scale (float): Scale parameter.
        loc (float): Location parameter (default 0).
    """

    def __init__(self, shape: float, scale: float, loc: float = 0) -> None:
        super().__init__(shape=shape, scale=scale, loc=loc)

    def cdf(self, x: NumberOrStochasticScalar) -> NumberOrStochasticScalar:
        shape, scale, loc = self._params.values()
        y = 1 / (1 + ((x - loc) / scale) ** shape)
        return 1 - y**shape

    def invcdf(self, u: NumberOrStochasticScalar) -> NumberOrStochasticScalar:
        shape, scale, loc = self._params.values()
        return loc + scale * (((1 - u) ** (-1 / shape)) - 1) ** (1 / shape)


class InverseBurr(DistributionBase):
    r"""Inverse Burr Distribution.

    Defined by:
        F(x) = [(( (x-μ)/σ )^τ / (1 + ((x-μ)/σ )^τ)]^α

    Parameters:
        power (float): Power parameter (τ).
        shape (float): Shape parameter (α).
        scale (float): Scale parameter (σ).
        loc (float): Location parameter (μ).
    """

    def __init__(self, power: float, shape: float, scale: float, loc: float) -> None:
        super().__init__(power=power, shape=shape, scale=scale, loc=loc)

    def cdf(self, x: NumberOrStochasticScalar) -> NumberOrStochasticScalar:
        power, shape, scale, loc = self._params.values()
        y = ((x - loc) / scale) ** power
        return (y / (1 + y)) ** shape

    def invcdf(self, u: NumberOrStochasticScalar) -> NumberOrStochasticScalar:
        power, shape, scale, loc = self._params.values()
        return (
            scale
            * (np.float_power((np.float_power(u, (-1 / shape)) - 1), (-1 / power)))
            + loc
        )


class InverseParalogistic(DistributionBase):
    r"""Inverse ParaLogistic Distribution.

    Represents an Inverse ParaLogistic distribution with given shape and scale parameters.
    Its CDF is defined as:

        F(x) = [(( (x-μ)/σ )^α / (1 + ((x-μ)/σ )^α)]^(-α),  x > μ

    Parameters:
        shape (float): The shape parameter.
        scale (float): The scale parameter.
        loc (float): The location parameter (default 0).
    """

    def __init__(self, shape: float, scale: float, loc: float = 0) -> None:
        super().__init__(shape=shape, scale=scale, loc=loc)

    def cdf(self, x: NumberOrStochasticScalar) -> NumberOrStochasticScalar:
        # Unpack parameters with explicit type annotations
        params = tuple(self._params.values())
        shape_val: float = params[0]
        scale_val: float = params[1]
        loc_val: float = params[2]
        y = ((x - loc_val) / scale_val) ** shape_val
        return (y / (1 + y)) ** shape_val

    def invcdf(self, u: NumberOrStochasticScalar) -> NumberOrStochasticScalar:
        params = tuple(self._params.values())
        shape_val: float = params[0]
        scale_val: float = params[1]
        loc_val: float = params[2]
        y = u ** (1 / shape_val)
        return loc_val + scale_val * (y / (1 - y)) ** (1 / shape_val)


class Weibull(DistributionBase):
    r"""Weibull Distribution.

    Defined by:
        F(x) = 1 - exp(-((x-μ)/σ)^α), x > μ

    Parameters:
        shape (float): Shape parameter (α).
        scale (float): Scale parameter (σ).
        loc (float): Location parameter (μ).
    """

    def __init__(self, shape: float, scale: float, loc: float = 0) -> None:
        super().__init__(shape=shape, scale=scale, loc=loc)

    def cdf(self, x: NumberOrStochasticScalar) -> NumberOrStochasticScalar:
        shape, scale, loc = self._params.values()
        y = ((x - loc) / scale) ** shape
        return -np.expm1(-y)

    def invcdf(self, u: NumberOrStochasticScalar) -> NumberOrStochasticScalar:
        shape, scale, loc = self._params.values()
        return loc + scale * (-np.log(1 - u)) ** (1 / shape)


class InverseWeibull(DistributionBase):
    r"""Inverse Weibull Distribution.

    Defined by:
        F(x) = exp(-((x-μ)/σ)^(-α)), x > μ

    Parameters:
        shape (float): Shape parameter (α).
        scale (float): Scale parameter (σ).
        loc (float): Location parameter (μ).
    """

    def __init__(self, shape: float, scale: float, loc: float = 0) -> None:
        super().__init__(shape=shape, scale=scale, loc=loc)

    def cdf(self, x: NumberOrStochasticScalar) -> NumberOrStochasticScalar:
        shape, scale, loc = self._params.values()
        y = np.float_power((x - loc) / scale, -shape)
        return np.exp(-y)

    def invcdf(self, u: NumberOrStochasticScalar) -> NumberOrStochasticScalar:
        shape, scale, loc = self._params.values()
        return loc + scale * (-1 / np.log(u)) ** (1 / shape)


class Exponential(DistributionBase):
    r"""Exponential Distribution.

    Defined by:
        F(x) = 1 - exp(-((x-μ)/σ)), x > μ

    Parameters:
        scale (float): Scale parameter.
        loc (float): Location parameter (default 0).
    """

    def __init__(self, scale: float, loc: float = 0) -> None:
        super().__init__(scale=scale, loc=loc)

    def cdf(self, x: NumberOrStochasticScalar) -> NumberOrStochasticScalar:
        scale, loc = self._params.values()
        y = (x - loc) / scale
        return -np.expm1(-y)

    def invcdf(self, u: NumberOrStochasticScalar) -> NumberOrStochasticScalar:
        scale, loc = self._params.values()
        return loc + scale * (-np.log(1 - u))


class Uniform(DistributionBase):
    r"""Uniform Distribution.

    Defined by:
        F(x) = (x - a) / (b - a), for a <= x <= b

    Parameters:
        a (float): Lower bound.
        b (float): Upper bound.
    """

    def __init__(self, a: float, b: float) -> None:
        super().__init__(a=a, b=b)

    def cdf(self, x: NumberOrStochasticScalar) -> NumberOrStochasticScalar:
        a, b = self._params.values()
        return (x - a) / (b - a)

    def invcdf(self, u: NumberOrStochasticScalar) -> NumberOrStochasticScalar:
        a, b = self._params.values()
        return a + (b - a) * u


class InverseExponential(DistributionBase):
    r"""Inverse Exponential Distribution.

    Defined by:
        F(x) = exp(-σ/(x-μ)), x > μ

    Parameters:
        scale (float): Scale parameter.
        loc (float): Location parameter (default 0).
    """

    def __init__(self, scale: float, loc: float = 0) -> None:
        super().__init__(scale=scale, loc=loc)

    def cdf(self, x: NumberOrStochasticScalar) -> NumberOrStochasticScalar:
        scale, loc = self._params.values()
        y = scale * np.float_power((x - loc), -1)
        return np.exp(-y)

    def invcdf(self, u: NumberOrStochasticScalar) -> NumberOrStochasticScalar:
        scale, loc = self._params.values()
        return loc - scale / np.log(u)


# --- Distribution Generator Classes ---

AVAILABLE_DISCRETE_DISTRIBUTIONS: Dict[str, Any] = {
    "poisson": Poisson,
    "negbinomial": NegBinomial,
    "binomial": Binomial,
    "hypergeometric": HyperGeometric,
}

AVAILABLE_CONTINUOUS_DISTRIBUTIONS: Dict[str, Any] = {
    "beta": Beta,
    "burr": Burr,
    "exponential": Exponential,
    "gamma": Gamma,
    "gpd": GPD,
    "logistic": Logistic,
    "lognormal": LogNormal,
    "loglogistic": LogLogistic,
    "normal": Normal,
    "paralogistic": Paralogistic,
    "pareto": Pareto,
    "uniform": Uniform,
    "inverseburr": InverseBurr,
    "inverseexponential": InverseExponential,
    "inversegamma": InverseGamma,
    "inverseparalogistic": InverseParalogistic,
    "inverseweibull": InverseWeibull,
    "weibull": Weibull,
}


class DistributionGeneratorBase(ABC):
    """
    Base class for parameterized distribution generators.
    Wraps a DistributionBase instance.
    """

    def __init__(self, distribution: DistributionBase) -> None:
        self.this_distribution = distribution

    def cdf(self, x: NumberOrStochasticScalar) -> NumberOrStochasticScalar:
        return self.this_distribution.cdf(x)

    def invcdf(self, u: NumberOrStochasticScalar) -> NumberOrStochasticScalar:
        return self.this_distribution.invcdf(u)

    def generate(
        self, n_sims: int | None = None, rng: np.random.Generator = config.rng
    ) -> StochasticScalar:
        return self.this_distribution.generate(n_sims, rng)


class DiscreteDistributionGenerator(DistributionGeneratorBase):
    """Discrete distribution generator instantiated by name."""

    def __init__(self, distribution_name: str, parameters: list) -> None:
        distribution_name = distribution_name.lower()
        if distribution_name not in AVAILABLE_DISCRETE_DISTRIBUTIONS:
            raise ValueError(
                f"Distribution {distribution_name} must be one of {list(AVAILABLE_DISCRETE_DISTRIBUTIONS.keys())}"
            )
        distribution_cls = AVAILABLE_DISCRETE_DISTRIBUTIONS[distribution_name]
        super().__init__(distribution_cls(*parameters))


class ContinuousDistributionGenerator(DistributionGeneratorBase):
    """Continuous distribution generator instantiated by name."""

    def __init__(self, distribution_name: str, parameters: list) -> None:
        distribution_name = distribution_name.lower()
        if distribution_name not in AVAILABLE_CONTINUOUS_DISTRIBUTIONS:
            raise ValueError(
                f"Distribution {distribution_name} must be one of {list(AVAILABLE_CONTINUOUS_DISTRIBUTIONS.keys())}"
            )
        distribution_cls = AVAILABLE_CONTINUOUS_DISTRIBUTIONS[distribution_name]
        super().__init__(distribution_cls(*parameters))



from variables import ProteusVariable

cumulative_triangle = [[20, 30, 40], [10, 20], [5]]
row_sums = [sum(row) for row in cumulative_triangle]
column_sums = [
    sum(row[i] for row in cumulative_triangle)
    for i in range(len(cumulative_triangle[0]))
]
development_factors = [
    [
        row_sums[j + 1] / (row_sums[j] - cumulative_triangle[j, j])
        for j in range(len(cumulative_triangle) - 1)
    ]
]



from __future__ import annotations
from .config import xp as np
from numpy.typing import ArrayLike
from .couplings import ProteusStochasticVariable, CouplingGroup
from typing import Union, TypeVar
import math
import plotly.graph_objects as go  # type: ignore

Numeric = Union[int, float]
NumberOrList = TypeVar("NumberOrList", Numeric, list[Numeric])
NumericOrStochasticScalar = TypeVar(
    "NumericOrStochasticScalar", Numeric, "StochasticScalar"
)


class StochasticScalar(ProteusStochasticVariable):
    """A class to represent a single scalar variable in a simulation."""

    coupled_variable_group: CouplingGroup

    @property
    def ranks(self) -> StochasticScalar:
        """Return the ranks of the variable."""
        result = np.empty(self.n_sims, dtype=int)
        result[np.argsort(self.values)] = np.arange(self.n_sims)
        return StochasticScalar(result)

    def __init__(self, values: ArrayLike):
        super().__init__()
        assert hasattr(values, "__getitem__"), "Values must be an array-like object."
        if isinstance(values, StochasticScalar):
            self.values = values.values
            self.n_sims = values.n_sims
            self.coupled_variable_group.merge(values.coupled_variable_group)
        else:
            if isinstance(values, list):
                self.values = np.array(values)
                self.n_sims = len(values)
            elif isinstance(values, np.ndarray):
                if values.ndim == 1:
                    self.values = values
                    self.n_sims = len(values)
                else:
                    raise ValueError("Values must be a 1D array.")
            else:
                raise ValueError(
                    "Values must be a list or numpy array. Found " + str(type(values))
                )

    def __hash__(self):
        return id(self)

    def tolist(self):
        return self.values.tolist()

    def _reorder_sims(self, new_order) -> None:
        """Reorder the simulations in the variable."""
        self.values = self.values[new_order]

    def __array_ufunc__(
        self, ufunc: np.ufunc, method: str, *inputs, **kwargs
    ) -> StochasticScalar:
        """Override the __array_ufunc__ method means that you can apply standard numpy functions"""
        # check if the input types to the function are types of ProteusVariables other than StochasticScalar
        var_not_stochastic_scalar = [
            isinstance(x, ProteusStochasticVariable)
            and not isinstance(x, StochasticScalar)
            for x in inputs
        ]

        if any(var_not_stochastic_scalar):
            # call the __array_ufunc__ method of variable which is not StochasticScalar
            #
            var_pos = var_not_stochastic_scalar.index(True)
            return inputs[var_pos].__array_ufunc__(ufunc, method, *inputs, **kwargs)
        _inputs = tuple(
            (
                x.values
                if isinstance(x, StochasticScalar)
                else x  # promote an input ndarray to match the simulation index
            )
            for x in inputs
        )
        out = kwargs.get("out", ())
        if out:
            kwargs["out"] = tuple(x.values for x in out)
        result = StochasticScalar(getattr(ufunc, method)(*_inputs, **kwargs))
        for input in inputs:
            if isinstance(input, ProteusStochasticVariable):
                input.coupled_variable_group.merge(self.coupled_variable_group)
        result.coupled_variable_group.merge(self.coupled_variable_group)

        return result

    def ssum(self) -> float:
        """Sum the values of the variable across the simulation dimension."""
        return np.sum(self.values)

    def mean(self) -> float:
        """Return the mean of the variable across the simulation dimension."""
        return np.mean(self.values)

    def skew(self) -> float:
        """Return the coefficient of skewness of the variable across the simulation dimension."""
        return float(np.mean((self.values - self.mean()) ** 3) / self.std() ** 3)

    def kurt(self) -> float:
        """Return the kurtosis of the variable across the simulation dimension."""
        return float(np.mean((self.values - self.mean()) ** 4) / self.std() ** 4)

    def std(self) -> float:
        """Return the standard deviation of the variable across the simulation dimension."""
        return np.std(self.values)

    def percentile(self, p: NumberOrList) -> NumberOrList:
        """Return the percentile of the variable across the simulation dimension."""
        return np.percentile(self.values, p)

    def tvar(self, p: NumberOrList) -> NumberOrList:
        """Return the tail value at risk (TVAR) of the variable."""
        # get the rank of the variable
        rank_positions = np.argsort(self.values)
        if isinstance(p, list):
            result = []
            for perc in p:
                result.append(
                    self.values[
                        rank_positions[math.ceil(perc / 100 * self.n_sims) :]
                    ].mean()
                )
            return result
        return self.values[rank_positions[math.ceil(p / 100 * self.n_sims) :]].mean()

    def upsample(self, n_sims: int) -> StochasticScalar:
        """Increase the number of simulations in the variable."""
        if n_sims == self.n_sims:
            return self
        return StochasticScalar(self.values[np.arange(n_sims) % self.n_sims])

    def __repr__(self):
        return f"StochasticScalar(values={self.values}\nn_sims={self.n_sims})"

    # implement the index referencing
    def __getitem__(
        self, index: NumericOrStochasticScalar
    ) -> NumericOrStochasticScalar:
        if isinstance(index, (int, float)):
            return self.values[int(index)]
        elif isinstance(index, StochasticScalar):
            result = StochasticScalar(self.values[index.values])
            result.coupled_variable_group.merge(index.coupled_variable_group)
            return result
        raise ValueError("Index must be an integer, StochasticScalar or numpy array.")

    def show_histogram(self, title: str | None = None):
        """Show a histogram of the variable.

        Args:
            title (str | None): Title of the histogram plot. Defaults to None.

        """
        fig = go.Figure(go.Histogram(x=self.values), layout=dict(title=title))
        fig.show()

    def show_cdf(self, title: str | None = None):
        """Show a plot of the cumulative distribution function (cdf) of the variable.

        Args:
            title (str | None): Title of the cdf plot. Defaults to None.

        """

        fig = go.Figure(
            go.Scatter(x=np.sort(self.values), y=np.arange(self.n_sims) / self.n_sims),
            layout=dict(title=title),
        )
        fig.update_xaxes(dict(title="Value"))
        fig.update_yaxes(dict(title="Cumulative Probability"))
        fig.show()



from typing import Union, Callable
import numpy as np
from .couplings import ProteusStochasticVariable
from .stochastic_scalar import (
    StochasticScalar,
)
from .config import config
from . import distributions

ProteusCompatibleTypes = Union["FreqSevSims", StochasticScalar, int, float, np.ndarray]


def _get_sims_of_events(n_events_by_sim: np.ndarray):
    """Given the number of events in each simulation, returns the simulation index for each event.

    >>> n_events_by_sim = np.array([1, 0, 3])
    >>> _get_sims_of_events(n_events_by_sim)
    array([0, 2, 2, 2])

    Parameters:
    - n_events_by_sim (np.ndarray): Array of the number of events in each simulation.

    Returns:
    - np.ndarray: Array of simulation indices for each event.
    """
    cumulative_n_events = n_events_by_sim.cumsum()
    total_events = cumulative_n_events[-1]
    event_no = np.arange(total_events)
    return cumulative_n_events.searchsorted(event_no + 1)


class FrequencySeverityModel:
    """A class for constructing and simulating from Frequency-Severity, or Compound distributions"""

    def __init__(
        self,
        freq_dist: distributions.DistributionBase,
        sev_dist: distributions.DistributionBase,
    ):
        self.freq_dist = freq_dist
        self.sev_dist = sev_dist

    def generate(
        self, n_sims=None, rng: np.random.Generator = config.rng
    ) -> "FreqSevSims":
        """
        Generate simulations from the Frequency-Severity model.

        Parameters:
        - n_sims (int): Number of simulations to generate. If None, uses the default value from the config.
        - rng (np.random.Generator): Random number generator. Defaults to the value from the config.

        Returns:
        - FreqSevSims: Object containing the generated simulations.
        """
        if n_sims is None:
            n_sims = config.n_sims
        n_events = self.freq_dist.generate(n_sims, rng)
        total_events = n_events.ssum()
        sev = self.sev_dist.generate(int(total_events), rng)
        result = FreqSevSims(_get_sims_of_events(n_events.values), sev.values, n_sims)
        result.coupled_variable_group.merge(n_events.coupled_variable_group)
        result.coupled_variable_group.merge(sev.coupled_variable_group)
        return result


class FreqSevSims(ProteusStochasticVariable):
    """A class for storing and manipulating Frequency-Severity simulations.
    FreqSevSims objects provide convenience methods for aggregating and summarizing the simulations.

    >>> sim_index = np.array([0, 0, 1, 1, 1, 2, 2, 2, 2])
    >>> values = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9])
    >>> n_sims = 3
    >>> fs = FreqSevSims(sim_index, values, n_sims)
    >>> fs.aggregate()
    StochasticScalar([ 3., 12., 30.])
    >>> fs.occurrence()
    StochasticScalar([2., 5., 9.])

    They can be operated on using standard mathematical operations, as well as as numpy ufuncs and functions.

    >>> fs + 1
    FreqSevSims(array([0, 0, 1, 1, 1, 2, 2, 2, 2]), array([ 2,  3,  4,  5,  6,  7,  8,  9, 10]), 3)
    >>> np.maximum(fs, 5)
    FreqSevSims(array([0, 0, 1, 1, 1, 2, 2, 2, 2]), array([5, 5, 5, 5, 5, 6, 7, 8, 9]), 3)
    >>> np.where(fs > 5, 1, 0)
    FreqSevSims(array([0, 0, 1, 1, 1, 2, 2, 2, 2]), array([0, 0, 0, 0, 0, 1, 1, 1, 1]), 3)

    FreqSevSims objects can be multiplied, added, subtracted, divided, and compared with other FreqSevSims objects,
    provided that the simulation indices match.

    >>> fs1 = FreqSevSims(sim_index, values, n_sims)
    >>> fs2 = FreqSevSims(sim_index, values, n_sims)
    >>> fs1 + fs2
    FreqSevSims(array([0, 0, 1, 1, 1, 2, 2, 2, 2]), array([ 2,  4,  6,  8, 10, 12, 14, 16, 18]), 3)
    """

    def __init__(
        self,
        sim_index: np.ndarray | list[int],
        values: np.ndarray | list[int],
        n_sims: int,
    ):
        """
        Create a new FreqSevSims object out the list of simulation indices, and the list of values corresponding to
        each simulation index. Note, the simulation indices are assumed to be ordered and 0-indexed.


        Parameters:
        sim_index (np.ndarray|list): Array of simulation indices.
        values (np.ndarray|list): Array of values.
        n_sims (int): Number of simulations.

        Raises:
        AssertionError: If lengths of values and sim_index don't match.


        """
        super().__init__()
        self.sim_index = np.asarray(sim_index)
        self.values = np.asarray(values)
        self.n_sims = n_sims

        assert len(self.sim_index) == len(self.values)

    def __hash__(self):
        return id(self)

    def __str__(self):
        return (
            "Simulation Index\n"
            + str(self.sim_index)
            + "\n Values\n"
            + str(self.values)
        )

    def _reorder_sims(self, ordering: np.ndarray) -> None:
        """Reorder the simulations of the FreqSevSims object according to the given order."""
        reverse_ordering = np.empty(len(ordering), dtype=int)
        reverse_ordering[ordering] = np.arange(len(ordering), dtype=int)
        self.sim_index = reverse_ordering[self.sim_index]

    def __getitem__(self, sim_index: int) -> StochasticScalar:
        """Returns the values of the simulation with the given simulation index."""
        # get the positions of the given simulation index
        if isinstance(sim_index, int):
            ints = np.where(self.sim_index == sim_index)
            return StochasticScalar(self.values[ints])
        else:
            raise NotImplementedError

    def _reduce_over_events(self, operation) -> StochasticScalar:
        _result = np.zeros(self.n_sims)
        operation(_result, self.sim_index, self.values)
        result = StochasticScalar(_result)
        result.coupled_variable_group.merge(self.coupled_variable_group)
        return result

    def aggregate(self) -> StochasticScalar:
        """Calculates the aggregate loss for each simulation.

        >>> sim_index = np.array([0, 0, 1, 1, 1, 2, 2, 2, 2])
        >>> values = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9])
        >>> n_sims = 3
        >>> fs = FreqSevSims(sim_index, values, n_sims)
        >>> fs.aggregate()
        array([ 3., 12., 30.])

        Returns:
            numpy.ndarray: An array containing the aggregate loss for each simulation.
        """
        return self._reduce_over_events(np.add.at)

    def occurrence(self) -> StochasticScalar:
        """Calculates the maximum occurrence loss for each simulation.

        >>> sim_index = np.array([0, 0, 1, 1, 1, 2, 2, 2, 2])
        >>> values = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9])
        >>> n_sims = 3
        >>> fs = FreqSevSims(sim_index, values, n_sims)
        >>> fs.occurrence()
        array([2., 5., 9.])

        Returns:
            numpy.ndarray: An array containing the aggregate loss for each simulation.
        """
        return self._reduce_over_events(np.maximum.at)

    def deep_copy(self) -> "FreqSevSims":
        """Creates a deep copy of the FreqSevSims object."""
        return FreqSevSims(self.sim_index, self.values.copy(), self.n_sims)

    def copy(self) -> "FreqSevSims":
        """Creates a copy of the FreqSevSims object."""
        result = FreqSevSims(self.sim_index, self.values.copy(), self.n_sims)
        result.coupled_variable_group.merge(self.coupled_variable_group)
        return result

    def apply(self, func) -> "FreqSevSims":
        """Applies a function to the values of the FreqSevSims object."""
        result = FreqSevSims(self.sim_index, func(self.values), self.n_sims)
        result.coupled_variable_group.merge(self.coupled_variable_group)
        return result

    def __array_ufunc__(
        self, ufunc: np.ufunc, method: str, *inputs, **kwargs
    ) -> "FreqSevSims":
        _inputs = tuple(
            (
                x.values
                if isinstance(x, FreqSevSims)
                else (
                    x[self.sim_index]
                    if isinstance(x, np.ndarray)
                    else (
                        x.values[self.sim_index]
                        if isinstance(x, StochasticScalar)
                        else x
                    )
                )  # promote an input ndarray to match the simulation index
            )
            for x in inputs
        )
        out = kwargs.get("out", ())
        if out:
            kwargs["out"] = tuple(x.values for x in out)
        result = getattr(ufunc, method)(*_inputs, **kwargs)
        result = FreqSevSims(self.sim_index, result, self.n_sims)
        for input in inputs:
            if isinstance(input, ProteusStochasticVariable):
                input.coupled_variable_group.merge(self.coupled_variable_group)
        result.coupled_variable_group.merge(self.coupled_variable_group)

        return result

    def __array_function__(self, func: Callable, types, args, kwargs):
        if func not in (np.where, np.sum):
            raise NotImplementedError
        args = tuple(x.values if isinstance(x, FreqSevSims) else x for x in args)
        result = func(*args, **kwargs)
        if isinstance(result, np.number):
            return result
        result = FreqSevSims(self.sim_index, result, self.n_sims)
        result.coupled_variable_group.merge(self.coupled_variable_group)
        return result

    def __repr__(self):
        return "%s(%r)" % (type(self).__name__, self.values)

    def _is_compatible(self, other: ProteusCompatibleTypes):
        """Check if two FreqSevSims objects are compatible for mathematical operations."""
        return isinstance(other, FreqSevSims) and self.sim_index is other.sim_index

    def upsample(self, n_sims: int) -> "FreqSevSims":
        """Upsamples the FreqSevSims object to the given number of simulations."""
        if n_sims == self.n_sims:
            return self.copy()
        sim_index = np.repeat(self.sim_index, n_sims // self.n_sims)
        values = np.repeat(self.values, n_sims // self.n_sims)
        if n_sims % self.n_sims > 0:
            sim_index = np.concatenate(
                (sim_index, self.sim_index[self.sim_index < n_sims % self.n_sims])
            )
            values = np.concatenate(
                (values, self.values[self.sim_index < n_sims % self.n_sims])
            )
        sim_index = sim_index + np.arange(len(sim_index)) % n_sims
        return FreqSevSims(sim_index, values, n_sims)



from pal.variables import StochasticScalar
from pal.couplings import ProteusStochasticVariable
from pal.config import xp as np
import pandas as pd


class SimEventId:
    """A class to represent a mapping of sim_no to event_id."""

    def __init__(self, sim_no, sim_event_id, n_sims: int):
        super().__init__()
        self.sim_no = sim_no
        self.values = sim_event_id
        self.reverse_mapping = {
            event_id: sim_no for sim_no, event_id in zip(sim_no, sim_event_id)
        }
        self.n_sims = n_sims

    def __getitem__(self, item):
        return self.values[item]

    def get_sim_no(self, event_id):
        return np.array(
            [self.reverse_mapping[e] for e in event_id]
        )  # need a faster map here

    def _reorder_sims(self, ordering):
        self.sim_no = ordering[self.sim_no]


class SimEventLossTable(ProteusStochasticVariable):

    def __init__(
        self, sim_event_id: SimEventId, loss, master_sim_event_table: SimEventId
    ):
        self.sim_event_id = sim_event_id
        self.loss = loss
        self.master_sim_event_table = master_sim_event_table
        self.n_sims = master_sim_event_table.n_sims

    @classmethod
    def from_dataframe(cls, df: pd.DataFrame, master_yet: SimEventId):
        return cls(
            SimEventId(
                master_yet.get_sim_no(df["SimEventId"].values),
                df["SimEventId"].values,
                n_sims=master_yet.n_sims,
            ),
            df["Loss"].values,
            master_yet,
        )

    def _reorder_sims(self, new_order):
        self.master_sim_event_table._reorder_sims(new_order)

    def __array_ufunc__(
        self, ufunc: np.ufunc, method: str, *inputs, **kwargs
    ) -> "SimEventLossTable":
        """Override the __array_ufunc__ method means that you can apply standard numpy functions"""
        inputs_list = list(
            (
                x.loss
                if isinstance(x, SimEventLossTable)
                else x  # promote an input ndarray to match the simulation index
            )
            for x in inputs
        )
        # if any of the inputs are StochasticScalars, we need to match the simulation index
        for i, x in enumerate(inputs_list):
            if isinstance(x, StochasticScalar):
                inputs_list[i] = x.values[self.sim_event_id.sim_no]
        out = kwargs.get("out", ())
        if out:
            kwargs["out"] = tuple(x.values for x in out)
        result = SimEventLossTable(
            self.sim_event_id,
            getattr(ufunc, method)(*inputs_list, **kwargs),
            self.master_sim_event_table,
        )

        return result

    def __add__(self, other):
        if isinstance(other, SimEventLossTable):
            # combine the two tables
            # look at the union of the event_ids, etc
            assert (
                self.master_sim_event_table == other.master_sim_event_table
            ), "Cannot add two tables with different master sim event tables"
            event_ids = np.unique(
                np.concatenate((self.sim_event_id.values, other.sim_event_id.values))
            )
            sim_nos = self.master_sim_event_table.get_sim_no(event_ids)
            # create a new table with the combined event_ids
            loss = np.zeros(len(event_ids))
            loss[np.searchsorted(event_ids, self.sim_event_id.values)] = self.loss
            np.add.at(
                loss, np.searchsorted(event_ids, other.sim_event_id.values), other.loss
            )
            return SimEventLossTable(
                SimEventId(sim_nos, event_ids, self.master_sim_event_table.n_sims),
                loss,
                master_sim_event_table=self.master_sim_event_table,
            )
        elif isinstance(other, float | int):
            return SimEventLossTable(
                self.sim_event_id, self.loss + other, self.master_sim_event_table
            )
        elif isinstance(other, StochasticScalar):
            return SimEventLossTable(
                self.sim_event_id,
                self.loss + other[self.sim_event_id.sim_no],
                self.master_sim_event_table,
            )
        else:
            raise ValueError(f"Cannot add SimEventLossTable to {type(other)} type")

    def __sub__(self, other):
        if isinstance(other, SimEventLossTable):
            # combine the two tables
            # look at the union of the event_ids, etc
            assert (
                self.master_sim_event_table == other.master_sim_event_table
            ), "Cannot add two tables with different master sim event tables"
            event_ids = np.unique(
                np.concatenate((self.sim_event_id.values, other.sim_event_id.values))
            )
            sim_nos = self.master_sim_event_table.get_sim_no(event_ids)
            # create a new table with the combined event_ids
            loss = np.zeros(len(event_ids))
            loss[np.searchsorted(event_ids, self.sim_event_id.values)] = self.loss
            np.subtract.at(
                loss, np.searchsorted(event_ids, other.sim_event_id.values), other.loss
            )
            return SimEventLossTable(
                SimEventId(sim_nos, event_ids, self.master_sim_event_table.n_sims),
                loss,
                master_sim_event_table=self.master_sim_event_table,
            )
        elif isinstance(other, float | int):
            return SimEventLossTable(
                self.sim_event_id, self.loss - other, self.master_sim_event_table
            )
        elif isinstance(other, StochasticScalar):
            return SimEventLossTable(
                self.sim_event_id,
                self.loss - other[self.sim_event_id.sim_no],
                self.master_sim_event_table,
            )
        else:
            raise ValueError(f"Cannot add SimEventLossTable to {type(other)} type")

    def __gt__(self, other):
        if isinstance(other, float | int):
            return SimEventLossTable(
                self.sim_event_id,
                self.loss > other,
                self.master_sim_event_table,
            )
        elif isinstance(other, StochasticScalar):
            return SimEventLossTable(
                self.sim_event_id,
                self.loss > other[self.sim_event_id.sim_no],
                self.master_sim_event_table,
            )
        else:
            raise ValueError(f"Cannot compare SimEventLossTable to {type(other)} type")

    def _rle__(self, other):
        return self.__gt__(other)

    def __le__(self, other):
        if isinstance(other, float | int):
            return SimEventLossTable(
                self.sim_event_id,
                self.loss <= other,
                self.master_sim_event_table,
            )
        else:
            raise ValueError(f"Cannot compare SimEventLossTable to {type(other)} type")

    def _rgt__(self, other):
        return self.__le__(other)

    def __radd__(self, other):
        return self.__add__(other)

    def __mul__(self, other):
        if isinstance(other, SimEventLossTable):
            # combine the two tables
            # look at the union of the event_ids, etc
            assert (
                self.master_sim_event_table == other.master_sim_event_table
            ), "Cannot add two tables with different master sim event tables"
            event_ids = np.union1d(self.sim_event_id.values, other.sim_event_id.values)
            sim_nos = self.master_sim_event_table.get_sim_no(event_ids)
            # create a new table with the combined event_ids
            loss = np.ones(len(event_ids))
            np.multiply.at(
                loss, np.searchsorted(event_ids, self.sim_event_id.values), self.loss
            )
            np.multiply.at(
                loss, np.searchsorted(event_ids, other.sim_event_id.values), other.loss
            )
            return SimEventLossTable(
                SimEventId(sim_nos, event_ids),
                loss,
                master_sim_event_table=self.master_sim_event_table,
            )
        elif isinstance(other, float | int):
            return SimEventLossTable(
                self.sim_event_id, self.loss * other, self.master_sim_event_table
            )
        elif isinstance(other, StochasticScalar):
            return SimEventLossTable(
                self.sim_event_id,
                self.loss * other[StochasticScalar(self.sim_event_id.sim_no)].values,
                self.master_sim_event_table,
            )
        else:
            return other.__rmul__(self)

    def __rmul__(self, other):
        return self.__mul__(other)

    def aggregate(self) -> StochasticScalar:
        """Aggregates the losses by sim_no."""
        result = np.zeros(self.n_sims)
        np.add.at(
            result,
            self.sim_event_id.sim_no,
            self.loss,
        )
        return StochasticScalar(result)

    def __repr__(self):
        return f"SimEventLossTable(\nSimEventIds{self.sim_event_id.values}\nLoss {self.loss})"



from pal import FreqSevSims
from .config import xp as np

percentiles = np.array([1, 2, 5, 10, 20, 50, 70, 80, 90, 95, 99, 99.5, 99.8, 99.9])


def loss_summary(losses: FreqSevSims):
    occurrence_losses = losses.occurrence()
    occurrence_statistics = np.percentile(occurrence_losses, percentiles)
    aggregate_losses = losses.aggregate()
    aggregate_statistics = np.percentile(aggregate_losses, percentiles)
    result = {"Occurrence": occurrence_statistics, "Aggregate": aggregate_statistics}
    return result



import scipy.stats  # ignore:import-untyped
from pal import copulas
from pal import distributions
import pytest
import numpy as np
import scipy
import scipy.special
from pal.variables import ProteusVariable, StochasticScalar


def copula_margins(copula_samples: list[StochasticScalar]):
    # check values are between 0 and 1
    y = ProteusVariable("dim1", [(x >= 0) & (x <= 1) for x in copula_samples])

    assert y.all()

    # check the values are uniform by checking the moments
    for u in copula_samples:
        assert np.isclose(u.mean(), 0.5, atol=1e-2)
        assert np.isclose(u.std(), 1 / np.sqrt(12), atol=1e-2)
        assert np.isclose(u.skew(), 0, atol=1e-1)
        assert np.isclose(u.kurt(), 3 - 6 / 5, atol=1e-1)


@pytest.mark.parametrize("correlation", [-0.999, 0.5, 0, -0.5, 0.25, 0.75, 0.999])
def test_gaussian_copula(correlation):
    samples = copulas.GaussianCopula([[1, correlation], [correlation, 1]]).generate(
        100000
    )
    # test the correlations
    emp_corr = np.corrcoef((samples[0].values, samples[1].values))[0, 1]
    # convert from rank to linear
    rank_corr = 2 * np.sin(emp_corr * np.pi / 6)
    assert np.isclose(rank_corr, correlation, atol=1e-2)
    k = scipy.stats.kendalltau(samples[0].values, samples[1].values).statistic
    assert np.isclose(k, 2 / np.pi * np.asin(correlation), atol=1e-2)
    # test the margins
    copula_margins(samples)


@pytest.mark.parametrize("correlation", [-0.999, 0.5, 0, -0.5, 0.25, 0.75, 0.999])
def test_gaussian_copula_apply(correlation):
    n_sims = 100000
    samples = [
        distributions.Gamma(2, 50).generate(n_sims),
        distributions.LogNormal(2, 1.5).generate(n_sims),
    ]
    copulas.GaussianCopula([[1, correlation], [correlation, 1]]).apply(samples)
    # test the correlations
    emp_corr = np.corrcoef((samples[0].ranks.values, samples[1].ranks.values))[0, 1]
    # convert from rank to linear
    linear_corr = 2 * np.sin(emp_corr * np.pi / 6)
    assert np.isclose(linear_corr, correlation, atol=1e-2)
    k = scipy.stats.kendalltau(samples[0].values, samples[1].values).statistic
    assert np.isclose(k, 2 / np.pi * np.asin(correlation), atol=1e-2)


@pytest.mark.parametrize("dof", [1.5, 5, 9, 100])
@pytest.mark.parametrize("correlation", [-0.999, 0.5, -0.5, 0, 0.25, 0.75, 0.999])
def test_studentst_copula(correlation, dof):
    samples = copulas.StudentsTCopula(
        [[1, correlation], [correlation, 1]], dof
    ).generate(100000)
    k = scipy.stats.kendalltau(samples[0].values, samples[1].values).statistic
    assert np.isclose(k, 2 / np.pi * np.asin(correlation), atol=1e-2)
    # test the margins
    copula_margins(samples)


@pytest.mark.parametrize("dof", [1.5, 5, 9, 100])
@pytest.mark.parametrize("correlation", [-0.999, 0.5, 0, -0.5, 0.25, 0.75, 0.999])
def test_studentst_copula_apply(correlation, dof):
    n_sims = 100000
    samples = [
        distributions.Gamma(2, 50).generate(n_sims),
        distributions.LogNormal(2, 1.5).generate(n_sims),
    ]
    copulas.StudentsTCopula([[1, correlation], [correlation, 1]], dof).apply(samples)
    # test the correlations
    k = scipy.stats.kendalltau(samples[0].values, samples[1].values).statistic
    assert np.isclose(k, 2 / np.pi * np.asin(correlation), atol=1e-2)


@pytest.mark.parametrize("alpha", [0.0, 0.5, 1.25, 2.75])
def test_clayton_copula(alpha):
    samples = copulas.ClaytonCopula(alpha, 2).generate(100000)
    k = scipy.stats.kendalltau(samples[0].values, samples[1].values).statistic
    assert np.isclose(k, alpha / (2 + alpha), atol=1e-2)
    # test the margins
    copula_margins(samples)


@pytest.mark.parametrize("alpha", [0.0, 0.5, 1.25, 2.75])
def test_clayton_copula_apply(alpha):
    n_sims = 100000
    samples = [
        distributions.Gamma(2, 50).generate(n_sims),
        distributions.LogNormal(2, 1.5).generate(n_sims),
    ]
    copulas.ClaytonCopula(alpha, 2).apply(samples)
    k = scipy.stats.kendalltau(samples[0].values, samples[1].values).statistic
    assert np.isclose(k, alpha / (2 + alpha), atol=1e-2)


@pytest.mark.parametrize("theta", [1.001, 1.25, 2.2, 5])
def test_gumbel_copula(theta):
    samples = copulas.GumbelCopula(theta, 2).generate(100000)
    # calculate the Kendall's tau value
    k = scipy.stats.kendalltau(samples[0].values, samples[1].values).statistic
    assert np.isclose(k, 1 - 1 / theta, atol=1e-2)
    # test the margins
    copula_margins(samples)


@pytest.mark.parametrize("theta", [1.001, 1.25, 2.2, 5])
def test_gumbel_copula_apply(theta):
    n_sims = 100000
    samples = [
        distributions.Gamma(2, 50).generate(n_sims),
        distributions.LogNormal(2, 1.5).generate(n_sims),
    ]
    copulas.GumbelCopula(theta, 2).apply(samples)
    k = scipy.stats.kendalltau(samples[0].values, samples[1].values).statistic
    assert np.isclose(k, 1 - 1 / theta, atol=1e-2)


@pytest.mark.parametrize("theta", [1.001, 1.25, 2.2, 3])
def test_joe_copula(theta):
    samples = copulas.JoeCopula(theta, 2).generate(100000)
    # calculate the Kendall's tau value
    k = scipy.stats.kendalltau(samples[0].values, samples[1].values).statistic
    assert np.isclose(
        k,
        1
        + 2
        / (2 - theta)
        * (scipy.special.digamma(2) - scipy.special.digamma(2 / theta + 1)),
        atol=1e-2,
    )
    # test the margins
    copula_margins(samples)


def debye1(x):
    """The first Debye function"""
    return (
        np.log(1 - np.exp(-x)) * x
        + scipy.special.zeta(2)
        - scipy.special.spence(1 - np.exp(-x))
    ) / x


@pytest.mark.parametrize("theta", [0.001, 0.5, 2, 4])
def test_frank_copula(theta):
    samples = copulas.FrankCopula(theta, 2).generate(100000)
    # calculate the Kendall's tau value
    k = scipy.stats.kendalltau(samples[0].values, samples[1].values).statistic
    assert np.isclose(
        k,
        1 + 4 / theta * (debye1(theta) - 1),
        atol=1e-2,
    )
    # test the margins
    copula_margins(samples)



from pal import FreqSevSims
import numpy as np
import pytest

x = FreqSevSims(np.array([0, 0, 0, 0]), np.array([100000, 800000, 500000, 200000]), 0)

print(x > 100000)

print(np.where(x > 100000, 0, 1))



"""
Copula Module

This module contains classes for representing and generating samples from various
copulas. It includes both elliptical (Gaussian and Student's T) and Archimedean
copulas.
"""

# Standard library imports
from abc import ABC, abstractmethod
from typing import Any, List, Tuple, Union

# Third-party imports
import numpy.typing as npt
import scipy.stats.distributions as distributions  # type: ignore

# Local imports
from .config import config, xp as np, _use_gpu
from .variables import ProteusVariable, StochasticScalar

if _use_gpu:
    import cupyx.scipy.special as special  # type: ignore
else:
    import scipy.special as special  # type: ignore


class Copula(ABC):
    """A base class to represent a copula."""

    @abstractmethod
    def generate(
        self, n_sims: Union[int, None] = None, rng: np.random.Generator = config.rng
    ) -> ProteusVariable:
        """Generate samples from the copula."""
        pass

    def _generate_unnormalised(
        self, n_sims: int, rng: np.random.Generator
    ) -> npt.NDArray:
        """
        Generate samples from the multivariate distribution underlying the copula.
        The marginal distribution of the samples will not necessarily be uniform.
        """
        return NotImplemented  # To be implemented by subclasses

    def apply(self, variables: Union[ProteusVariable, List[StochasticScalar]]) -> None:
        """
        Apply the copula to a list of variables.

        Parameters:
            variables: A ProteusVariable or list of StochasticScalar instances.
        """
        # Generate the copula samples
        copula_samples = ProteusVariable(
            dim_name="dim1",
            values=[
                StochasticScalar(sample)
                for sample in self._generate_unnormalised(
                    n_sims=variables[0].n_sims, rng=config.rng
                )
            ],
        )
        if len(variables) != len(copula_samples):
            raise ValueError("Number of variables and copula samples do not match.")
        # Apply the copula to the variables
        apply_copula(variables, copula_samples)


class EllipticalCopula(Copula, ABC):
    """A base class to represent an elliptical copula."""

    matrix: npt.NDArray
    chol: npt.NDArray

    def __init__(
        self,
        matrix: npt.NDArray | list[list[float]],
        *args: Any,
        matrix_type: str = "linear",
        **kwargs: Any,
    ) -> None:
        _matrix = np.asarray(matrix)
        if _matrix.ndim != 2 or _matrix.shape[0] != _matrix.shape[1]:
            raise ValueError("Matrix must be square")
        if matrix_type == "linear":
            self.correlation_matrix = _matrix
            # Check that the correlation matrix is positive definite
            try:
                self.chol = np.linalg.cholesky(self.correlation_matrix)
            except np.linalg.LinAlgError:
                raise ValueError("Correlation matrix is not positive definite")
        elif matrix_type == "chol":
            self.chol = _matrix
        else:
            raise ValueError("matrix_type must be 'linear' or 'chol'")
        self.matrix = _matrix


class GaussianCopula(EllipticalCopula):
    """A class to represent a Gaussian copula."""

    def __init__(self, matrix: npt.NDArray, matrix_type: str = "linear") -> None:
        super().__init__(matrix, matrix_type=matrix_type)

    def generate(
        self, n_sims: Union[int, None] = None, rng: np.random.Generator = config.rng
    ) -> ProteusVariable:
        """Generate samples from the Gaussian copula."""
        if n_sims is None:
            n_sims = config.n_sims

        # Generate samples from a multivariate normal distribution
        samples = self._generate_unnormalised(n_sims, rng)
        uniform_samples = special.ndtr(samples)
        result = ProteusVariable(
            "dim1", [StochasticScalar(sample) for sample in uniform_samples]
        )
        for val in result:
            val.coupled_variable_group.merge(result[0].coupled_variable_group)
        return result

    def _generate_unnormalised(
        self, n_sims: int, rng: np.random.Generator
    ) -> npt.NDArray:
        n_vars = self.correlation_matrix.shape[0]
        normal_samples = rng.multivariate_normal(
            mean=np.zeros(n_vars), cov=np.eye(n_vars), size=n_sims
        ).T
        return self.chol.dot(normal_samples)


class StudentsTCopula(EllipticalCopula):
    """A class to represent a Student's T copula."""

    def __init__(
        self,
        matrix: npt.NDArray[np.float64] | list[list[float]],
        dof: float,
        matrix_type: str = "linear",
    ) -> None:
        super().__init__(matrix, matrix_type=matrix_type)
        if dof <= 0:
            raise ValueError("Degrees of Freedom must be positive")
        self.dof = dof

    def generate(
        self, n_sims: Union[int, None] = None, rng: np.random.Generator = config.rng
    ) -> ProteusVariable:
        if n_sims is None:
            n_sims = config.n_sims
        t_samples = self._generate_unnormalised(n_sims, rng)
        uniform_samples = distributions.t(self.dof).cdf(t_samples)
        return ProteusVariable(
            "dim1", [StochasticScalar(sample) for sample in uniform_samples]
        )

    def _generate_unnormalised(
        self, n_sims: int, rng: np.random.Generator
    ) -> npt.NDArray:
        n_vars = self.correlation_matrix.shape[0]
        normal_samples = self.chol.dot(
            rng.multivariate_normal(
                mean=np.zeros(n_vars), cov=np.eye(n_vars), size=n_sims
            ).T
        )
        chi_samples = np.sqrt(rng.gamma(self.dof / 2, 2 / self.dof, size=n_sims))
        return normal_samples / chi_samples[np.newaxis, :]


class ArchimedeanCopula(Copula, ABC):
    """A base class to represent an Archimedean copula."""

    @abstractmethod
    def generator_inv(self, t: npt.NDArray) -> npt.NDArray:
        """The inverse generator function of the copula."""
        pass

    @abstractmethod
    def generate_latent_distribution(
        self, n_sims: int, rng: np.random.Generator
    ) -> npt.NDArray:
        """Generate samples from the latent distribution of the copula."""
        pass

    def __init__(self, n: int) -> None:
        self.n = n

    def generate(
        self, n_sims: Union[int, None] = None, rng: np.random.Generator = config.rng
    ) -> ProteusVariable:
        copula_samples = self.generator_inv(-self._generate_unnormalised(n_sims, rng))
        result = ProteusVariable(
            "dim1", [StochasticScalar(sample) for sample in copula_samples]
        )
        for val in result:
            val.coupled_variable_group.merge(result[0].coupled_variable_group)
        return result

    def _generate_unnormalised(
        self, n_sims: Union[int, None] = None, rng: np.random.Generator = config.rng
    ) -> npt.NDArray:
        if n_sims is None:
            n_sims = config.n_sims
        n_vars = self.n
        # Generate samples from a uniform distribution
        u = rng.uniform(size=(n_vars, n_sims))
        # Generate samples from the latent distribution
        latent_samples = self.generate_latent_distribution(n_sims, rng)
        # Calculate the copula samples
        return np.log(u) / latent_samples[np.newaxis]


class ClaytonCopula(ArchimedeanCopula):
    """A class to represent a Clayton copula."""

    def __init__(self, theta: float, n: int) -> None:
        if theta < 0:
            raise ValueError("Theta cannot be negative")
        self.theta = theta
        self.n = n

    def generator_inv(self, t: npt.NDArray) -> npt.NDArray:
        if self.theta == 0:
            return np.exp(-t)
        return (1 + t) ** (-1 / self.theta)

    def generate_latent_distribution(
        self, n_sims: int, rng: np.random.Generator
    ) -> npt.NDArray:
        if self.theta == 0:
            return np.array([1])
        return rng.gamma(1 / self.theta, size=n_sims)


def levy_stable(
    alpha: float,
    beta: float,
    size: Union[int, Tuple[int, ...]],
    rng: np.random.Generator,
) -> npt.NDArray:
    """
    Simulate samples from a Lévy stable distribution using the Chambers-Mallows-Stuck algorithm.

    Parameters:
        alpha (float): Stability parameter in (0, 2].
        beta (float): Skewness parameter in [-1, 1].
        size (int or tuple of ints): Output shape.
        rng (np.random.Generator): Random number generator.

    Returns:
        np.ndarray: Samples from the Lévy stable distribution.
    """
    U = rng.uniform(-np.pi / 2, np.pi / 2, size)
    W = rng.exponential(1, size)

    if alpha != 1:
        theta = np.arctan(beta * np.tan(np.pi * alpha / 2)) / alpha
        factor = (1 + beta**2 * np.tan(np.pi * alpha / 2) ** 2) ** (1 / (2 * alpha))
        part1 = np.sin(alpha * (U + theta)) / (np.cos(U)) ** (1 / alpha)
        part2 = (np.cos(U - alpha * (U + theta)) / W) ** ((1 - alpha) / alpha)
        samples = factor * part1 * part2
    else:
        samples = (2 / np.pi) * (
            (np.pi / 2 + beta * U) * np.tan(U)
            - beta * np.log((np.pi / 2 * W * np.cos(U)) / (np.pi / 2 + beta * U))
        )
    return samples


class GumbelCopula(ArchimedeanCopula):
    """A class to represent a Gumbel copula."""

    def __init__(self, theta: float, n: int) -> None:
        if theta < 1:
            raise ValueError("Theta must be at least 1")
        self.theta = theta
        self.n = n

    def generator_inv(self, t: npt.NDArray) -> npt.NDArray:
        return np.exp(-(t ** (1 / self.theta)))

    def generate_latent_distribution(
        self, n_sims: int, rng: np.random.Generator
    ) -> npt.NDArray:
        return levy_stable(1 / self.theta, 1, n_sims, rng) * (
            np.cos(np.pi / (2 * self.theta)) ** self.theta
        )


class FrankCopula(ArchimedeanCopula):
    """A class to represent a Frank copula."""

    def __init__(self, theta: float, n: int) -> None:
        self.theta = theta
        self.n = n

    def generator_inv(self, t: npt.NDArray) -> npt.NDArray:
        return -np.log1p(np.exp(-t) * (np.expm1(-self.theta))) / self.theta

    def generate_latent_distribution(
        self, n_sims: int, rng: np.random.Generator
    ) -> npt.NDArray:
        return rng.logseries(1 - np.exp(-self.theta), size=n_sims)


class JoeCopula(ArchimedeanCopula):
    """A class to represent a Joe copula."""

    def __init__(self, theta: float, n: int) -> None:
        if theta < 1:
            raise ValueError("Theta must be in the range [1, inf)")
        self.theta = theta
        self.n = n

    def generator_inv(self, t: npt.NDArray) -> npt.NDArray:
        return 1 - (1 - np.exp(-t)) ** (1 / self.theta)

    def generate_latent_distribution(
        self, n_sims: int, rng: np.random.Generator
    ) -> npt.NDArray:
        return _sibuya_gen(1 / self.theta, n_sims, rng)


def _sibuya_gen(
    alpha: float, size: Union[int, Tuple[int, ...]], rng: np.random.Generator
) -> npt.NDArray:
    """
    Generate samples from a Sibuya distribution.

    Parameters:
        alpha (float): Parameter for the Sibuya distribution.
        size (int or tuple): Output shape.
        rng (np.random.Generator): Random number generator.

    Returns:
        np.ndarray: Samples from a Sibuya distribution.
    """
    g1 = rng.gamma(alpha, 1, size=size)
    g2 = rng.gamma(1 - alpha, 1, size=size)
    r = g2 / g1
    e = rng.exponential(1, size=size)
    u = r * e
    return 1 + rng.poisson(u, size=size)


def apply_copula(
    variables: Union[ProteusVariable, List[StochasticScalar]],
    copula_samples: Union[ProteusVariable, List[StochasticScalar]],
) -> None:
    """
    Apply a reordering from a copula to a list of variables.

    Parameters:
        variables: List of StochasticScalar variables.
        copula_samples: List of StochasticScalar samples from the copula.
    """
    if len(variables) != len(copula_samples):
        raise ValueError("Number of variables and copula samples do not match.")
    variables_list: list[StochasticScalar] = []
    for var in variables:
        assert isinstance(
            var, StochasticScalar
        ), "Variables must be StochasticScalar instances"
        variables_list.append(var)
    for i, var1 in enumerate(variables_list):
        for j, var2 in enumerate(variables_list[i + 1 :]):
            if var1.coupled_variable_group is var2.coupled_variable_group:
                raise ValueError(
                    f"Cannot apply copula as the variables at positions {i} and {j + i + 1} are not independent"
                )
    copula_sort_indices = np.argsort(
        np.array([cs.values for cs in copula_samples]), axis=1, kind="stable"
    )
    copula_ranks = np.argsort(copula_sort_indices, axis=1)
    variable_sort_indices = np.argsort(
        np.array([var.values for var in variables]), axis=1
    )
    first_variable_rank = np.argsort(variable_sort_indices[0])
    copula_ranks = copula_ranks[:, copula_sort_indices[0, first_variable_rank]]
    for i, var in enumerate(variables):
        if i == 0:
            continue
        re_ordering = variable_sort_indices[i, copula_ranks[i]]
        for var2 in var.coupled_variable_group.variables:
            var2._reorder_sims(re_ordering)
    for var in variables:
        var.coupled_variable_group.merge(variables[0].coupled_variable_group)



from pal.variables import StochasticScalar
from pal.copulas import apply_copula, GumbelCopula
from pal.frequency_severity import FreqSevSims, FrequencySeverityModel
from pal.distributions import GPD, Poisson, Normal
import numpy as np
import scipy


def test_fs_reordering():
    x = FreqSevSims([0, 0, 1, 2, 4], [10, 21, 30, 40, 50], 6)
    y = FreqSevSims([0, 1, 1, 3, 5], [12, 22, 32, 42, 52], 6)
    a = x.aggregate()  # a is [31,30,40,0,50,0]
    b = y.aggregate()  # b is [12,54,0,42,0,52]
    copula_samples = [
        StochasticScalar([0, 2, 1, 5, 4, 3]),
        StochasticScalar([1, 0, 3, 2, 5, 4]),
    ]
    apply_copula([a, b], copula_samples)
    assert (a.values == [31, 30, 40, 0, 50, 0]).all()
    assert (b.values == [52, 0, 54, 0, 12, 42]).all()
    # expected_x
    expected_x = FreqSevSims([0, 0, 1, 2, 4], [10, 21, 30, 40, 50], 6)
    # expected_y
    expected_y = FreqSevSims([4, 2, 2, 5, 0], [12, 22, 32, 42, 52], 6)

    assert (expected_x.sim_index == x.sim_index).all()
    assert (expected_x.values == x.values).all()
    assert (expected_y.sim_index == y.sim_index).all()
    assert (expected_y.values == y.values).all()


def test_fs_reordering2():
    """Tests that other variables attached to the FreqSevSims are correctly reordered."""
    x = FreqSevSims([0, 0, 1, 2, 4], [10, 21, 30, 40, 50], 6)
    y = FreqSevSims([0, 1, 1, 3, 5], [12, 22, 32, 42, 52], 6)
    x1 = x * 2
    y1 = y * 3
    a = x1.aggregate()  # a is [31,30,40,0,50,0]*2
    b = y1.aggregate()  # b is [12,54,0,42,0,52]*3
    copula_samples = [
        StochasticScalar([0, 2, 1, 5, 4, 3]),
        StochasticScalar([1, 0, 3, 2, 5, 4]),
    ]
    apply_copula([a, b], copula_samples)
    assert (a.values == [62, 60, 80, 0, 100, 0]).all()
    assert (b.values == [156, 0, 162, 0, 36, 126]).all()
    # expected_x
    expected_x = FreqSevSims([0, 0, 1, 2, 4], [10, 21, 30, 40, 50], 6)
    # expected_y
    expected_y = FreqSevSims([4, 2, 2, 5, 0], [12, 22, 32, 42, 52], 6)

    assert (expected_x.sim_index == x.sim_index).all()
    assert (expected_x.values == x.values).all()
    assert (expected_y.sim_index == y.sim_index).all()
    assert (expected_y.values == y.values).all()

    # aggregate x
    x_agg = x.aggregate()
    y_agg = y.aggregate()
    expected_x_agg = StochasticScalar([31, 30, 40, 0, 50, 0])
    expected_y_agg = StochasticScalar([52, 0, 54, 0, 12, 42])
    assert (x_agg == expected_x_agg).values.all()
    assert (y_agg == expected_y_agg).values.all()


def test_fs_reordering3():
    """Tests that other variables attached to the FreqSevSims are correctly reordered."""
    x = FrequencySeverityModel(
        Poisson(mean=2), GPD(shape=0.33, scale=100000, loc=0)
    ).generate()
    y = FrequencySeverityModel(
        Poisson(mean=2), GPD(shape=0.33, scale=100000, loc=0)
    ).generate()
    x1 = x * 2
    y1 = y * 3
    a = x1.aggregate()
    b = y1.aggregate()
    GumbelCopula(1.5, 2).apply([a, b])
    # check the copula has been applied correctly
    calculated_tau = scipy.stats.kendalltau(a.values, b.values).statistic
    assert np.isclose(calculated_tau, 1 - 1 / 1.5, atol=1e-2)
    # check that when x and y are re_calculated and reaggregated, they give the same result
    re_calculated_a = (x * 2).aggregate()
    re_calculated_b = (y * 3).aggregate()
    assert np.allclose(re_calculated_a.values, a.values, atol=1e-10)
    assert np.allclose(re_calculated_b.values, b.values, atol=1e-10)



import numpy as np
import pytest
from pal.frequency_severity import FreqSevSims
from pal.variables import StochasticScalar


def test_aggregate():
    sim_index = np.array([0, 0, 1, 1, 1, 2, 2, 2, 2])
    values = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9])
    n_sims = 3
    fs = FreqSevSims(sim_index, values, n_sims)
    assert np.array_equal(fs.aggregate().values, np.array([3.0, 12.0, 30.0]))


def test_occurrence():
    sim_index = np.array([0, 0, 1, 1, 1, 2, 2, 2, 2])
    values = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9])
    n_sims = 3
    fs = FreqSevSims(sim_index, values, n_sims)
    assert np.array_equal(fs.occurrence().values, np.array([2.0, 5.0, 9.0]))


def test_copy():
    sim_index = np.array([0, 0, 1, 1, 1, 2, 2, 2, 2])
    values = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9])
    n_sims = 3
    fs = FreqSevSims(sim_index, values, n_sims)
    fs_copy = fs.copy()
    assert np.array_equal(fs_copy.sim_index, fs.sim_index)
    assert np.array_equal(fs_copy.values, fs.values)
    assert fs_copy.n_sims == fs.n_sims


def test_apply():
    sim_index = np.array([0, 0, 1, 1, 1, 2, 2, 2, 2])
    values = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9])
    n_sims = 3
    fs = FreqSevSims(sim_index, values, n_sims)
    fs_squared = fs.apply(lambda x: x**2)
    assert np.array_equal(
        fs_squared.values, np.array([1, 4, 9, 16, 25, 36, 49, 64, 81])
    )


def test_math_operations():
    sim_index = np.array([0, 0, 1, 1, 1, 2, 2, 2, 2])
    values = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9])
    n_sims = 3
    fs1 = FreqSevSims(sim_index, values, n_sims)
    fs2 = FreqSevSims(sim_index, values, n_sims)
    fs_sum = fs1 + fs2
    assert np.array_equal(fs_sum.values, np.array([2, 4, 6, 8, 10, 12, 14, 16, 18]))
    fs_plus_one = fs1 + 1
    assert np.array_equal(fs_plus_one.values, np.array([2, 3, 4, 5, 6, 7, 8, 9, 10]))
    fs_times_five = fs1 * 5
    assert np.array_equal(
        fs_times_five.values, np.array([5, 10, 15, 20, 25, 30, 35, 40, 45])
    )
    fs1_less1 = -1 + fs1
    assert np.array_equal(fs1_less1.values, np.array([0, 1, 2, 3, 4, 5, 6, 7, 8]))
    fs_divide2 = fs1 / 2
    assert np.array_equal(
        fs_divide2.values, np.array([0.5, 1, 1.5, 2, 2.5, 3, 3.5, 4, 4.5])
    )
    fs_squared = fs1**2
    assert np.array_equal(
        fs_squared.values, np.array([1, 4, 9, 16, 25, 36, 49, 64, 81])
    )
    fs1_times_fs2 = fs1 * fs2
    assert np.array_equal(
        fs1_times_fs2.values, np.array([1, 4, 9, 16, 25, 36, 49, 64, 81])
    )
    fs1_divide_fs2 = fs1 / fs2
    assert np.array_equal(fs1_divide_fs2.values, np.array([1, 1, 1, 1, 1, 1, 1, 1, 1]))
    two_to_power_fs1 = 2**fs1
    assert np.array_equal(
        two_to_power_fs1.values, np.array([2, 4, 8, 16, 32, 64, 128, 256, 512])
    )
    # test with numpy array
    arr = np.array([1, 2, 3])
    fs1_times_arr = fs1 * arr
    assert np.array_equal(
        fs1_times_arr.values, np.array([1, 2, 6, 8, 10, 18, 21, 24, 27])
    )
    fs1_divide_arr = fs1 / arr
    assert np.array_equal(
        fs1_divide_arr.values, np.array([1, 2, 1.5, 2, 2.5, 2, 7 / 3, 8 / 3, 3])
    )
    fs1_plus_arr = fs1 + arr
    assert np.array_equal(fs1_plus_arr.values, np.array([2, 3, 5, 6, 7, 9, 10, 11, 12]))
    fs1_minus_arr = fs1 - arr
    assert np.array_equal(fs1_minus_arr.values, np.array([0, 1, 1, 2, 3, 3, 4, 5, 6]))
    # test reverse operations
    fs1_times_arr = arr * fs1
    assert np.array_equal(
        fs1_times_arr.values, np.array([1, 2, 6, 8, 10, 18, 21, 24, 27])
    )
    fs1_divide_arr = arr / fs1
    assert np.array_equal(
        fs1_divide_arr.values,
        np.array([1, 0.5, 2 / 3, 0.5, 0.4, 0.5, 3 / 7, 0.375, 1 / 3]),
    )
    fs1_plus_arr = arr + fs1
    assert np.array_equal(fs1_plus_arr.values, np.array([2, 3, 5, 6, 7, 9, 10, 11, 12]))
    fs1_minus_arr = arr - fs1
    assert np.array_equal(
        fs1_minus_arr.values, np.array([0, -1, -1, -2, -3, -3, -4, -5, -6])
    )
    fs1_to_power_arr = fs1**arr
    assert np.array_equal(
        fs1_to_power_arr.values, np.array([1, 2, 9, 16, 25, 216, 343, 512, 729])
    )
    arr_to_power_fs1 = arr**fs1
    assert np.array_equal(
        arr_to_power_fs1.values, np.array([1, 1, 8, 16, 32, 729, 2187, 6561, 19683])
    )


def test_get_sims():
    sim_index = np.array([0, 0, 1, 1, 1, 2, 2, 2, 2])
    values = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9])
    fs = FreqSevSims(sim_index=sim_index, values=values, n_sims=3)
    selected_values = fs[1]
    assert (selected_values.values == [3, 4, 5]).all()


def test_stochastic_scalar():
    sim_index = np.array([0, 0, 1, 1, 1, 2, 2, 2, 2])
    values = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9])
    fs = FreqSevSims(sim_index=sim_index, values=values, n_sims=3)
    scalar = StochasticScalar([1, 2, 3])
    result = fs + scalar
    assert np.array_equal(result.values, np.array([2, 3, 5, 6, 7, 9, 10, 11, 12]))
    assert (
        result.coupled_variable_group
        == fs.coupled_variable_group
        == scalar.coupled_variable_group
    )
    fs = FreqSevSims(sim_index=sim_index, values=values, n_sims=3)
    scalar = StochasticScalar([1, 2, 3])
    result = fs - scalar
    assert (
        result.coupled_variable_group
        == fs.coupled_variable_group
        == scalar.coupled_variable_group
    )
    assert np.array_equal(result.values, np.array([0, 1, 1, 2, 3, 3, 4, 5, 6]))
    fs = FreqSevSims(sim_index=sim_index, values=values, n_sims=3)
    scalar = StochasticScalar([1, 2, 3])
    result = fs * scalar
    assert (
        result.coupled_variable_group
        == fs.coupled_variable_group
        == scalar.coupled_variable_group
    )
    assert np.array_equal(result.values, np.array([1, 2, 6, 8, 10, 18, 21, 24, 27]))
    fs = FreqSevSims(sim_index=sim_index, values=values, n_sims=3)
    scalar = StochasticScalar([1, 2, 3])
    result = fs / scalar
    assert np.array_equal(
        result.values, np.array([1, 2, 1.5, 2, 2.5, 2, 2 + 1 / 3, 2 + 2 / 3, 3])
    )
    assert (
        result.coupled_variable_group
        == fs.coupled_variable_group
        == scalar.coupled_variable_group
    )
    fs = FreqSevSims(sim_index=sim_index, values=values, n_sims=3)
    scalar = StochasticScalar([1, 2, 3])
    result = fs**scalar
    assert np.array_equal(
        result.values, np.array([1, 2, 9, 16, 25, 216, 343, 512, 729])
    )
    assert (
        result.coupled_variable_group
        == fs.coupled_variable_group
        == scalar.coupled_variable_group
    )
    fs = FreqSevSims(sim_index=sim_index, values=values, n_sims=3)
    result = fs + 1
    assert np.array_equal(result.values, np.array([2, 3, 4, 5, 6, 7, 8, 9, 10]))
    fs = FreqSevSims(sim_index=sim_index, values=values, n_sims=3)
    result = fs - 1
    assert np.array_equal(result.values, np.array([0, 1, 2, 3, 4, 5, 6, 7, 8]))

    # reverse operations
    fs = FreqSevSims(sim_index=sim_index, values=values, n_sims=3)
    scalar = StochasticScalar([1, 2, 3])
    result = scalar + fs
    assert np.array_equal(result.values, np.array([2, 3, 5, 6, 7, 9, 10, 11, 12]))
    assert (
        result.coupled_variable_group
        == fs.coupled_variable_group
        == scalar.coupled_variable_group
    )

    # reverse operations
    fs = FreqSevSims(sim_index=sim_index, values=values, n_sims=3)
    scalar = 1
    result = scalar + fs
    assert np.array_equal(result.values, np.array([2, 3, 4, 5, 6, 7, 8, 9, 10]))
    assert result.coupled_variable_group == fs.coupled_variable_group


def test_where():
    sim_index = np.array([0, 0, 1, 1, 1, 2, 2, 2, 2])
    values = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9])
    n_sims = 3
    fs = FreqSevSims(sim_index=sim_index, values=values, n_sims=n_sims)
    condition = fs > 5
    result = np.where(condition, fs, 0)
    assert np.array_equal(
        result.values, np.array([0.0, 0.0, 0.0, 0.0, 0.0, 6.0, 7, 8, 9])
    )
    assert result.n_sims == fs.n_sims
    assert (
        condition.coupled_variable_group
        == fs.coupled_variable_group
        == result.coupled_variable_group
    )


def test_minimum():
    sim_index = np.array([0, 0, 1, 1, 1, 2, 2, 2, 2])
    values = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9])
    n_sims = 3
    fs = FreqSevSims(sim_index=sim_index, values=values, n_sims=n_sims)
    result = np.minimum(fs, 5)
    assert np.array_equal(result.values, np.array([1, 2, 3, 4, 5, 5, 5, 5, 5]))
    assert result.n_sims == fs.n_sims
    assert result.coupled_variable_group == fs.coupled_variable_group


def test_maximum():
    sim_index = np.array([0, 0, 1, 1, 1, 2, 2, 2, 2])
    values = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9])
    n_sims = 3
    fs = FreqSevSims(sim_index=sim_index, values=values, n_sims=n_sims)
    result = np.maximum(fs, 5)
    assert np.array_equal(result.values, np.array([5, 5, 5, 5, 5, 6, 7, 8, 9]))
    assert result.n_sims == fs.n_sims
    assert result.coupled_variable_group == fs.coupled_variable_group



import numpy as np

from pal.variables import StochasticScalar
import pytest  # noqa


def test_empty():
    """Tests an empty stochastic scalar."""
    x = StochasticScalar([])
    assert x.n_sims == 0
    assert x.ssum() == 0


def test_stochastic_scalar():
    """Test that stochastic scalars can be created and manipulated."""
    x = StochasticScalar([4, 5, 2, 1, 3])
    assert (x.values == [4, 5, 2, 1, 3]).all()
    assert x.ssum() == 15
    assert x.mean() == 3
    # assert x.variance() == 2
    assert x.std() == 2**0.5
    assert x.percentile(50) == 3
    assert (x.percentile([10, 90]) == [1.4, 4.6]).all()
    assert (x + 1 == StochasticScalar([5, 6, 3, 2, 4])).values.all()


def test_add():
    """Tests the addition of two stochastic scalars."""
    x = StochasticScalar([4, 5, 2, 1, 3])
    y = StochasticScalar([1, 2, 3, 4, 5])
    z = x + y
    assert (z.values == [5, 7, 5, 5, 8]).all()
    assert (
        x.coupled_variable_group == y.coupled_variable_group == z.coupled_variable_group
    )


def test_add_self():
    """Tests the addition of a stochastic scalar to itself."""
    x = StochasticScalar([4, 5, 2, 1, 3])
    z = x + x
    assert (z.values == [8, 10, 4, 2, 6]).all()
    assert x.coupled_variable_group == z.coupled_variable_group


def test_add_self_inplace():
    """Tests the inplace addition of two stochastic scalars."""
    y = StochasticScalar(np.zeros(5))
    x = StochasticScalar([4, 5, 2, 1, 3])
    for i in range(10):
        y += x
    assert (y.values == [40, 50, 20, 10, 30]).all()


def test_add_scalar():
    """Tests the addition of a stochastic scalar and a scalar."""
    x = StochasticScalar([4, 5, 2, 1, 3])
    z = x + 1
    assert (z.values == [5, 6, 3, 2, 4]).all()
    assert x.coupled_variable_group == z.coupled_variable_group


def test_radd_scalar():
    """Tests the addition of a scalar and a stochastic scalar."""
    x = StochasticScalar([4, 5, 2, 1, 3])
    z = 1 + x
    assert (z.values == [5, 6, 3, 2, 4]).all()


def test_subtract():
    """Tests the subtraction of two stochastic scalars."""
    x = StochasticScalar([4, 5, 2, 1, 3])
    y = StochasticScalar([1, 2, 3, 4, 5])
    z = x - y
    assert (z.values == [3, 3, -1, -3, -2]).all()
    assert (
        x.coupled_variable_group == y.coupled_variable_group == z.coupled_variable_group
    )


def test_subtract_scalar():
    """Tests the subtraction of a stochastic scalar and a scalar."""
    x = StochasticScalar([4, 5, 2, 1, 3])
    z = x - 1
    assert (z.values == [3, 4, 1, 0, 2]).all()
    assert x.coupled_variable_group == z.coupled_variable_group


def test_rsubtract_scalar():
    """Tests the subtraction of a scalar and a stochastic scalar."""
    x = StochasticScalar([4, 5, 2, 1, 3])
    z = 1 - x
    assert (z.values == [-3, -4, -1, 0, -2]).all()
    assert x.coupled_variable_group == z.coupled_variable_group


def test_multiply():
    """Tests the multiplication of two stochastic scalars."""
    x = StochasticScalar([4, 5, 2, 1, 3])
    y = StochasticScalar([1, 2, 3, 4, 5])
    z = x * y
    assert (z.values == [4, 10, 6, 4, 15]).all()
    assert (
        x.coupled_variable_group == y.coupled_variable_group == z.coupled_variable_group
    )


def test_multiply_scalar():
    """Tests the multiplication of a stochastic scalar and a scalar."""
    x = StochasticScalar([4, 5, 2, 1, 3])
    z = x * 2
    assert (z.values == [8, 10, 4, 2, 6]).all()
    assert x.coupled_variable_group == z.coupled_variable_group


def test_rmultiply_scalar():
    """Tests the multiplication of a scalar and a stochastic scalar."""
    x = StochasticScalar([4, 5, 2, 1, 3])
    z = 2 * x
    assert (z.values == [8, 10, 4, 2, 6]).all()
    assert x.coupled_variable_group == z.coupled_variable_group


def test_divide():
    """Tests the division of two stochastic scalars."""
    x = StochasticScalar([4, 5, 2, 1, 3])
    y = StochasticScalar([1, 2, 3, 4, 5])
    z = x / y
    assert (z.values == [4, 2.5, 2 / 3, 0.25, 0.6]).all()
    assert (
        x.coupled_variable_group == y.coupled_variable_group == z.coupled_variable_group
    )


def test_divide_scalar():
    """Tests the division of a stochastic scalar and a scalar."""
    x = StochasticScalar([4, 5, 2, 1, 3])
    z = x / 2
    assert (z.values == [2, 2.5, 1, 0.5, 1.5]).all()
    assert x.coupled_variable_group == z.coupled_variable_group


def test_rdivide_scalar():
    """Tests the division of a scalar and a stochastic scalar."""
    x = StochasticScalar([4, 5, 2, 1, 3])
    z = 2 / x
    assert (z.values == [0.5, 0.4, 1, 2, 2 / 3]).all()
    assert x.coupled_variable_group == z.coupled_variable_group


def test_power():
    """Tests the power of two stochastic scalars."""
    x = StochasticScalar([4, 5, 2, 1, 3])
    y = StochasticScalar([1, 2, 3, 4, 5])
    z = x**y
    assert (z.values == [4, 25, 8, 1, 243]).all()
    assert (
        x.coupled_variable_group == y.coupled_variable_group == z.coupled_variable_group
    )


def test_power_scalar():
    """Tests the power of a stochastic scalar and a scalar."""
    x = StochasticScalar([4, 5, 2, 1, 3])
    z = x**2
    assert (z.values == [16, 25, 4, 1, 9]).all()
    assert x.coupled_variable_group == z.coupled_variable_group


def test_rpower_scalar():
    """Tests the power of a scalar and a stochastic scalar."""
    x = StochasticScalar([4, 5, 2, 1, 3])
    z = 2**x
    assert (z.values == [16, 32, 4, 2, 8]).all()
    assert x.coupled_variable_group == z.coupled_variable_group


def test_eq():
    """Tests the logical equals of two stochastic scalars."""
    x = StochasticScalar([4, 5, 2, 1, 3])
    y = StochasticScalar([4, 2, 3, 1, 5])
    z = x == y
    assert (z.values == [True, False, False, True, False]).all()
    assert (
        x.coupled_variable_group == y.coupled_variable_group == z.coupled_variable_group
    )


def test_eq_scalar():
    """Tests the logical equals of two stochastic scalars."""
    x = StochasticScalar([4, 5, 2, 1, 3])
    y = 1
    z = x == y
    assert (z.values == [False, False, False, True, False]).all()
    assert x.coupled_variable_group == z.coupled_variable_group


def test_not_eq():
    """Tests the logical not equals of two stochastic scalars."""
    x = StochasticScalar([4, 5, 2, 1, 3])
    y = StochasticScalar([4, 2, 3, 1, 5])
    z = x != y
    assert (z.values == [False, True, True, False, True]).all()
    assert (
        x.coupled_variable_group == y.coupled_variable_group == z.coupled_variable_group
    )


def test_not_eq_scalar():
    """Tests the logical not equals of two stochastic scalars."""
    x = StochasticScalar([4, 5, 2, 1, 3])
    y = 1
    z = x != y
    assert (z.values == [True, True, True, False, True]).all()
    assert x.coupled_variable_group == z.coupled_variable_group


def test_and():
    """Tests the logical and of two stochastic scalars."""
    x = StochasticScalar([False, True, False, True, True])
    y = StochasticScalar([False, False, True, True, True])
    z = x & y
    assert (z.values == [False, False, False, True, True]).all()
    assert (
        x.coupled_variable_group == y.coupled_variable_group == z.coupled_variable_group
    )


def test_numpy_ufunc():
    """Tests that a numpy ufunc can be applied to a stochastic scalar."""
    x = StochasticScalar([4, 5, 2, 1, 3])
    y = np.exp(x)
    assert (y.values == np.exp([4, 5, 2, 1, 3])).all()
    assert x.coupled_variable_group == y.coupled_variable_group
    assert type(y) is StochasticScalar


def test_ssum():
    """Tests the sum of a stochastic scalar."""
    x = StochasticScalar([4, 5, 2, 1, 3])
    y = x.ssum()
    assert y == 15


def test_mean():
    """Tests the mean of a stochastic scalar."""
    x = StochasticScalar([4, 5, 2, 1, 3])
    y = x.mean()
    assert y == 3


def test_dereference():
    x = StochasticScalar([4, 5, 2, 1, 3])
    y = x[3]
    assert y == 1


def test_stochastic_dereference():
    x = StochasticScalar([4, 5, 2, 1, 3])
    inds = StochasticScalar([3, 2, 1, 0, 0, 4])
    y = x[inds]
    assert (y.values == [1, 2, 5, 4, 4, 3]).all()
    assert inds.coupled_variable_group == y.coupled_variable_group


def test_percentile():
    x = StochasticScalar([4, 5, 2, 1, 3])
    y = x.percentile(50)
    assert y == 3
    y = x.percentile([10, 90])
    assert (y == [1.4, 4.6]).all()


def test_tvar():
    """Test the tail value at risk (TVAR) of the variable."""
    x = StochasticScalar([4, 5, 2, 1, 3])
    y = x.tvar(50)
    assert y == 4.5


def test_tvar2():
    """Test the tail value at risk (TVAR) of the variable."""
    x = StochasticScalar([4, 5, 2, 1, 3])
    y = x.tvar([50, 80])
    assert y == [4.5, 5]
